#!/usr/bin/env python
#-*- coding:utf-8 -*-

import numpy as np

import shapely
from shapely.affinity import scale
from shapely.geometry import Point, Polygon

from .dxf_import import DXF


'''
Shape generation from DXF files.
'''


__all__ = ["polygons_from_dxf"]


def polygons_from_dxf(filename, interpolate_curve=50, parent=None,
                      return_points=False):
    '''
    Generate :class:`shapely.geometry.Polygon` objects from a DXF file.
    '''
    dxf = DXF(filename, interpolate_curve=interpolate_curve)
    shapes = dxf.shapes()
    if return_points:
        points = {'path': []}
        for shape in shapes:
            points['path'].append(np.array(shape.exterior.coords))
        return shapes, points
    return shapes



#-*- coding:utf-8 -*-
#
# io/graph_loading.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Loading functions """

import ast
import codecs
import logging
import pickle
import types

import numpy as np

import nngt
from nngt.lib import InvalidArgument
from nngt.lib.logger import _log_message
from ..geometry import Shape, _shapely_support
from .io_helpers import _get_format
from .loading_helpers import *


logger = logging.getLogger(__name__)


# ---------- #
# Formatting #
# ---------- #

di_get_edges = {
    "neighbour": _get_edges_neighbour,
    "edge_list": _get_edges_elist,
    "gml": _get_edges_gml,
}


# ------------- #
# Load function #
# ------------- #

def load_from_file(filename, fmt="auto", separator=" ", secondary=";",
                   attributes=None, attributes_types=None, notifier="@",
                   ignore="#", name="LoadedGraph", directed=True,
                   cleanup=False):
    '''
    Load a Graph from a file.

    .. versionchanged :: 2.0
        Added optional `attributes_types` and `cleanup` arguments.

    .. warning ::
       Support for GraphML and DOT formats are currently limited and require
       one of the non-default backends (DOT requires graph-tool).

    Parameters
    ----------
    filename: str
        The path to the file.
    fmt : str, optional (default: "neighbour")
        The format used to save the graph. Supported formats are: "neighbour"
        (neighbour list, default if format cannot be deduced automatically),
        "ssp" (scipy.sparse), "edge_list" (list of all the edges in the graph,
        one edge per line, represented by a ``source target``-pair), "gml"
        (gml format, default if `filename` ends with '.gml'), "graphml"
        (graphml format, default if `filename` ends with '.graphml' or '.xml'),
        "dot" (dot format, default if `filename` ends with '.dot'), "gt" (only
        when using `graph_tool`<http://graph-tool.skewed.de/>_ as library,
        detected if `filename` ends with '.gt').
    separator : str, optional (default " ")
        separator used to separate inputs in the case of custom formats (namely
        "neighbour" and "edge_list")
    secondary : str, optional (default: ";")
        Secondary separator used to separate attributes in the case of custom
        formats.
    attributes : list, optional (default: [])
        List of names for the attributes present in the file. If a `notifier`
        is present in the file, names will be deduced from it; otherwise the
        attributes will be numbered.
        For "edge_list", attributes may also be present as additional columns
        after the source and the target.
    attributes_types : dict, optional (default: str)
        Backup information if the type of the attributes is not specified
        in the file. Values must be callables (types or functions) that will
        take the argument value as a string input and convert it to the proper
        type.
    notifier : str, optional (default: "@")
        Symbol specifying the following as meaningfull information. Relevant
        information are formatted ``@info_name=info_value``, where
        ``info_name`` is in ("attributes", "directed", "name", "size") and
        associated ``info_value`` are of type (``list``, ``bool``, ``str``,
        ``int``).
        Additional notifiers are ``@type=SpatialGraph/Network/SpatialNetwork``,
        which must be followed by the relevant notifiers among ``@shape``,
        ``@structure``, and ``@graph``.
    ignore : str, optional (default: "#")
        Ignore lines starting with the `ignore` string.
    name : str, optional (default: from file information or 'LoadedGraph')
        The name of the graph.
    directed : bool, optional (default: from file information or True)
        Whether the graph is directed or not.
    cleanup : bool, optional (default: False)
       If true, removes nodes before the first one that appears in the
       edges and after the last one and renumber the nodes from 0.

    Returns
    -------
    graph : :class:`~nngt.Graph` or subclass
        Loaded graph.
    '''
    return nngt.Graph.from_file(
        filename, fmt=fmt, separator=separator, secondary=secondary,
        attributes=attributes, attributes_types=attributes_types,
        notifier=notifier, ignore=ignore, name=name, directed=directed,
        cleanup=cleanup) 


def _load_from_file(filename, fmt="auto", separator=" ", secondary=";",
                    attributes=None, attributes_types=None,
                    notifier="@", ignore="#", cleanup=False):
    '''
    Load the main properties (edges, attributes...) from a file.

    Parameters
    ----------
    filename: str
        The path to the file.
    fmt : str, optional (default: "neighbour")
        The format used to save the graph. Supported formats are: "neighbour"
        (neighbour list, default if format cannot be deduced automatically),
        "ssp" (scipy.sparse), "edge_list" (list of all the edges in the graph,
        one edge per line, represented by a ``source target``-pair), "gml"
        (gml format, default if `filename` ends with '.gml'), "graphml"
        (graphml format, default if `filename` ends with '.graphml' or '.xml'),
        "dot" (dot format, default if `filename` ends with '.dot'), "gt" (only
        when using `graph_tool <http://graph-tool.skewed.de/>`_ as library,
        detected if `filename` ends with '.gt').
    separator : str, optional (default " ")
        separator used to separate inputs in the case of custom formats (namely
        "neighbour" and "edge_list")
    secondary : str, optional (default: ";")
        Secondary separator used to separate attributes in the case of custom
        formats.
    attributes : list, optional (default: [])
        List of names for the edge attributes present in the file. If a
        `notifier` is present in the file, names will be deduced from it;
        otherwise the attributes will be numbered.
    attributes_types : dict, optional (default: str)
        Backup information if the type of the attributes is not specified
        in the file. Values must be callables (types or functions) that will
        take the argument value as a string input and convert it to the proper
        type.
    notifier : str, optional (default: "@")
        Symbol specifying the following as meaningfull information. Relevant
        information are formatted ``@info_name=info_value``, where
        ``info_name`` is in ("attributes", "directed", "name", "size") and
        associated ``info_value`` are of type (``list``, ``bool``, ``str``,
        ``int``).
        Additional notifiers are ``@type=SpatialGraph/Network/SpatialNetwork``,
        which must be followed by the relevant notifiers among ``@shape``,
        ``@structure``, and ``@graph``.
    ignore : str, optional (default: "#")
        Ignore lines starting with the `ignore` string.
    cleanup : bool, optional (default: False)
       If true, removes nodes before the first one that appears in the
       edges and after the last one and renumber the nodes from 0. 

    Returns
    -------
    di_notif : dict
        Dictionary containing the main graph arguments.
    edges : list of 2-tuples
        Edges of the graph.
    di_nattributes : dict
        Dictionary containing the node attributes.
    di_eattributes : dict
        Dictionary containing the edge attributes (name as key, value as a
        list sorted in the same order as `edges`).
    struct : :class:`~nngt.NeuralPop`
        Population (``None`` if not present in the file).
    shape : :class:`~nngt.geometry.Shape`
        Shape of the graph (``None`` if not present in the file).
    positions : array-like of shape (N, d)
        The positions of the neurons (``None`` if not present in the file).
    '''
    # check for mpi
    if nngt.get_config("mpi"):
        raise NotImplementedError("This function is not ready for MPI yet.")

    # load
    lst_lines, struct, shape, positions = None, None, None, None
    fmt = _get_format(fmt, filename)

    if fmt not in ("neighbour", "edge_list", "gml"):
        return [None]*7

    with open(filename, "r") as filegraph:
        lst_lines = _process_file(filegraph, fmt, separator)

    # notifier lines
    di_notif = _get_notif(lst_lines, notifier, attributes, fmt=fmt,
                          atypes=attributes_types)

    # get nodes attributes
    di_nattributes = _get_node_attr(di_notif, separator, fmt=fmt,
                                    lines=lst_lines, atypes=attributes_types)

    # make edges and attributes
    eattributes     = di_notif["edge_attributes"]
    di_eattributes  = {name: [] for name in eattributes}
    di_edge_convert = _gen_convert(di_notif["edge_attributes"],
                                   di_notif["edge_attr_types"],
                                   attributes_types=attributes_types)

    # process file
    edges = di_get_edges[fmt](
        lst_lines, eattributes, ignore, notifier, separator, secondary,
        di_attributes=di_eattributes, di_convert=di_edge_convert,
        di_notif=di_notif)

    if cleanup:
        edges = np.array(edges) - np.min(edges)

    # add missing size information if necessary
    if "size" not in di_notif:
        di_notif["size"] = int(np.max(edges)) + 1

    # check whether a shape is present
    if 'shape' in di_notif:
        if _shapely_support:
            min_x, max_x = float(di_notif['min_x']), float(di_notif['max_x'])
            unit = di_notif['unit']
            shape = Shape.from_wkt(
                di_notif['shape'], min_x=min_x, max_x=max_x, unit=unit)
            # load areas
            try:
                def_areas      = ast.literal_eval(di_notif['default_areas'])
                def_areas_prop = ast.literal_eval(
                    di_notif['default_areas_prop'])

                for k in def_areas:
                    p = {key: float(v) for key, v in def_areas_prop[k].items()}
                    if "default_area" in k:
                        shape._areas["default_area"]._prop.update(p)
                        shape._areas["default_area"].height = p["height"]
                    else:
                        a = Shape.from_wkt(def_areas[k], unit=unit)
                        shape.add_area(a, height=p["height"], name=k,
                                       properties=p)

                ndef_areas      = ast.literal_eval(
                                      di_notif['non_default_areas'])
                ndef_areas_prop = ast.literal_eval(
                                      di_notif['non_default_areas_prop'])
                for i in ndef_areas:
                    p = {k: float(v) for k, v in ndef_areas_prop[i].items()}
                    a = Shape.from_wkt(ndef_areas[i], unit=unit)
                    shape.add_area(a, height=p["height"], name=i, properties=p)
            except KeyError:
                # backup compatibility with older versions
                pass
        else:
            _log_message(logger, "WARNING",
                         'A Shape object was present in the file but could '
                         'not be loaded because Shapely is not installed.')

    # check whether a structure is present
    if 'structure' in di_notif:
        str_enc = di_notif['structure'].replace('~', '\n').encode()
        str_dec = codecs.decode(str_enc, "base64")
        try:
            struct = pickle.loads(str_dec)
        except UnicodeError:
            struct = pickle.loads(str_dec, encoding="latin1")

    if 'x' in di_notif:
        x = np.fromstring(di_notif['x'], sep=separator)
        y = np.fromstring(di_notif['y'], sep=separator)
        if 'z' in di_notif:
            z = np.fromstring(di_notif['z'], sep=separator)
            positions = np.array((x, y, z)).T
        else:
            positions = np.array((x, y)).T

    return (di_notif, edges, di_nattributes, di_eattributes, struct, shape,
            positions)


def _library_load(filename, fmt):
    ''' Load the file using the library functions '''
    if nngt.get_config("backend") == "networkx":
        import networkx as nx

        if fmt == "graphml":
            return nx.read_graphml(filename)
        else:
            raise NotImplementedError
    elif nngt.get_config("backend") == "igraph":
        import igraph as ig

        if fmt == "graphml":
            return ig.Graph.Read_GraphML(filename)
        else:
            raise NotImplementedError
    elif nngt.get_config("backend") == "graph-tool":
        import graph_tool as gt

        return gt.load_graph(filename, fmt=fmt)
    else:
        raise NotImplementedError



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2020  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Layouts for topological representations
=======================================
"""

import os

import matplotlib as mpl
import matplotlib.pyplot as plt

import nngt


plt.rcParams.update({
    "figure.facecolor": (0, 0, 0, 0),
    "axes.labelcolor": "grey", "axes.titlecolor": "grey", "text.color": "grey"
})


nngt.seed(0)


# set matplotlib backend depending on the library
mpl_backend = mpl.get_backend()

if nngt.get_config("backend") in ("graph-tool", "igraph"):

    if mpl_backend.startswith("Qt4"):
        if mpl_backend != "Qt4Cairo":
            plt.switch_backend("Qt4Cairo")
    elif mpl_backend.startswith("Qt5"):
        if mpl_backend != "Qt5Cairo":
            plt.switch_backend("Qt5Cairo")
    elif mpl_backend.startswith("GTK"):
        if mpl_backend != "GTK3Cairo":
            plt.switch_backend("GTK3Cairo")
    elif mpl_backend != "cairo":
        plt.switch_backend("cairo")


# prepare figure and parameters

_, axes = plt.subplots(2, 2, figsize=(10, 8))

num_nodes = 50


# spring-block layout for structured graph

room1 = nngt.Group(10)
room2 = nngt.Group(20)
room3 = nngt.Group(20)

names = ["R1", "R2", "R3"]

struct = nngt.Structure.from_groups((room1, room2, room3), names)

g = nngt.Graph(structure=struct)

for room in struct:
    nngt.generation.connect_groups(g, room, room, "erdos_renyi", avg_deg=5)

nngt.generation.connect_groups(g, (room1, room2), struct, "erdos_renyi",
                               avg_deg=3, ignore_invalid=True)

nngt.generation.connect_groups(g, room3, room1, "erdos_renyi", avg_deg=5)

nngt.plot.library_draw(g, tight=False, axis=axes[0, 0], ecolor="grey",
                       show=False)

axes[0, 0].set_title("Spring-block layout")


# random layout

sw = nngt.generation.watts_strogatz(4, 0.3, nodes=num_nodes)

betw = nngt.analysis.betweenness(sw, "node")

nngt.plot.draw_network(sw, nsize=betw, ncolor="out-degree", axis=axes[0, 1],
                       ecolor="lightgrey", tight=False, show=False)

axes[0, 1].set_title("Random layout")


# circular layout for small-world networks

nngt.plot.draw_network(sw, nsize=betw, ncolor="out-degree", layout="circular",
                       ecolor="lightgrey", axis=axes[1, 0], show=False, tight=False)

axes[1, 0].set_title("Circular layout")


# spatial layout

c1 = nngt.geometry.Shape.disk(100)
c2 = nngt.geometry.Shape.disk(100, centroid=(50, 0))

shape = nngt.geometry.Shape.from_polygon(c1.union(c2))

npos  = shape.seed_neurons(num_nodes)

g = nngt.generation.distance_rule(10, shape=shape, nodes=num_nodes, avg_deg=5)

cc = nngt.analysis.local_clustering(g)

nngt.plot.draw_network(g, ncolor=cc, axis=axes[1, 1], ecolor="lightgrey",
                       tight=False, show=False)

axes[1, 1].set_title("Spatial layout")

plt.tight_layout()

# save figure

fname = os.getcwd() + "/layouts.png"

plt.savefig(fname)
plt.switch_backend(mpl_backend)

img = plt.imread(fname)

_, ax = plt.subplots(figsize=(10, 8))
ax.imshow(img)

ax.axis('off')

plt.tight_layout()
plt.show()

try:
    os.remove(fname)
except:
    pass

# %%
# Note that the last lines are just a little trick to make the figure be
# automatically detected by Sphinx-gallery. For normal use cases you can just
# do a regular ``plt.show()``.




#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# pync_log.py
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Logging functions for PyNCulture with checks for logging with MPI
'''

# MPI checks

def on_master_process():
    '''
    Check whether the current code is executing on the master process (rank 0)
    if MPI is used.

    Returns
    -------
    True if rank is 0, if mpi4py is not present or if MPI is not used,
    otherwise False.
    '''
    try:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0:
            return True
        else:
            return False
    except ImportError:
        return True


def mpi_checker(func):
    '''
    Decorator used to check for mpi and make sure only rank zero is used
    to store and generate the graph if the mpi algorithms are activated.
    '''
    def wrapper(*args, **kwargs):
        if on_master_process():
            return func(*args,**kwargs)
        else:
            return None
    return wrapper


# Logging messages
@mpi_checker
def _log_message(logger, level, message):
    if level == 'DEBUG':
        logger.debug(message)
    elif level == 'WARNING':
        logger.warning(message)
    elif level == 'INFO':
        logger.info(message)
    else:
        logger.critical(message)



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2020  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Geospatial networks
===================
"""

import os

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

import nngt
import nngt.geospatial as ng


plt.rcParams.update({
    'axes.edgecolor': 'grey', 'xtick.color': 'grey', 'ytick.color': 'grey',
    "figure.facecolor": (0, 0, 0, 0), "axes.facecolor": (0, 0, 0, 0),
    "axes.labelcolor": "grey", "axes.titlecolor": "grey", "text.color": "grey"
})


nngt.seed(2)


# take random countries
num_nodes = 20

world = ng.maps["adaptive"]
units = nngt._rng.choice(50, num_nodes, replace=False)
codes = list(world.iloc[units].SU_A3)

# make random network
g = nngt.generation.erdos_renyi(nodes=num_nodes, avg_deg=3)

# add the A3 code for each country (that's the crucial part that will link
# the graph to the geospatial data)
g.new_node_attribute("code", "string", codes)

g.set_weights(nngt._rng.exponential(2, g.edge_nb()))

# plot using draw_map and the A3 codes stored in "code"
ng.draw_map(g, "code", ncolor="in-degree", esize="weight", threshold=0,
            ecolor="grey", proj=ccrs.EqualEarth(), show=False)

plt.tight_layout()
plt.show()



#-*- coding:utf-8 -*-
#
# lib/converters.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Loading helpers """

import types

import numpy as np


def _to_np_array(data, dtype):
    '''
    Transform data to a numpy array, preserving nested lists as lists if dtype
    is "object".
    '''
    dtype = object if dtype == "string" else dtype

    if dtype in (object, "object"):
        # preserve potential list entries
        arr    = np.empty(len(data), dtype=object)
        arr[:] = data
        return arr

    return np.array(data, dtype=dtype)


def _to_list(string):
    separators = (';', ',', ' ', '\t')
    count = -np.Inf
    chosen = -1
    for i, delim in enumerate(separators):
        current = string.count(delim)
        if count < current:
            count = current
            chosen = separators[i]
            break
    return string.split(chosen)


def _to_int(string):
    try:
        return int(string)
    except ValueError:
        return int(float(string))


def _to_string(byte_string):
    ''' Convert bytes to string '''
    if isinstance(byte_string, bytes):
        return str(byte_string.decode())

    return byte_string


def _string_from_object(obj):
    ''' Return a type string from an object (usually a class) '''
    if obj.__class__ == type:
        # dealing with a class
        if issubclass(obj, float):
            return "double"

        if issubclass(obj, (int, np.integer)):
            return "int"

        if issubclass(obj, str):
            return "string"

        return "object"
    elif issubclass(obj.__class__, (types.FunctionType, types.MethodType)):
        # dealing with a function
        return "object"

    raise ValueError("Cannot deduce class string from '{}'.".format(obj))


def _np_dtype(attribute_type):
    '''
    Return a relevant numpy dtype entry.
    '''
    if attribute_type in ("double", "float", "real"):
        return float
    elif attribute_type in ("int", "integer"):
        return int

    return object


def _type_converter(attribute_type):
    if not isinstance(attribute_type, str):
        return attribute_type

    return _np_dtype(attribute_type)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# test_analysis.py
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
==============
Testing module
==============

This module tests the various functionalities of NNGT to make sure that all
implementations remain compatible with the graph libraries and versions 2.7 and
3.x of python.

note ::
    When adding new tests, filename should be of the form `test_xxx.py` and the
    code should contain:
    * a ``TestXXX`` class,
    * a ``suite = unittest.TestLoader().loadTestsFromTestCase(TestXXX)``
      declaration.
"""

# std imports
import sys
import importlib
from os import listdir, environ
from os.path import abspath, dirname, isfile
import unittest

# personal library
import nngt


# ------------- #
# Get the tests #
# ------------- #

# get the arguments for the graph library
backend = environ.get("GL", None)

if backend == "gt":
    nngt.use_backend("graph-tool")
    assert nngt.get_config('backend') == "graph-tool", \
           "Loading graph-tool failed..."
elif backend == "ig":
    nngt.use_backend("igraph")
    assert nngt.get_config('backend') == "igraph", \
           "Loading igraph failed..."
elif backend == "nx":
    nngt.use_backend("networkx")
    assert nngt.get_config('backend') == "networkx", \
           "Loading networkx failed..."
elif backend == "nngt":
    nngt.use_backend("nngt")
    assert nngt.get_config('backend') == "nngt", \
           "Loading nngt failed..."


# get the arguments for MPI/OpenMP + hide log
omp = int(environ.get("OMP", 1))
mpi = bool(environ.get("MPI", False))

conf = {
    "multithreading": omp > 1,
    "omp": omp,
    "mpi": mpi,
    "log_level": "ERROR",
}

nngt.set_config(conf, silent=True)


# get the tests
current_dir = dirname(abspath(__file__))
dir_files = listdir(current_dir)
sys.path.insert(0, current_dir)
testfiles = [fname[:-3] for fname in dir_files if (fname.startswith("test_")
             and fname.endswith(".py"))]

# remove the MPI test unless we're using it, otherwise remove the examples
if not nngt.get_config("mpi"):
    idx = None
    for i, test in enumerate(testfiles):
        if "test_mpi" in test:
            idx = i
            break
    del testfiles[idx]
else:
    idx = None
    for i, test in enumerate(testfiles):
        if "test_examples" in test:
            idx = i
            break
    del testfiles[idx]


tests = [importlib.import_module(name) for name in testfiles]


# ----------- #
# Run if main #
# ----------- #

if __name__ == "__main__":
    for test in tests:
        if hasattr(test, "suite"):
            unittest.TextTestRunner(verbosity=2).run(test.suite)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    import pint
    from pint import UnitRegistry, set_application_registry

    # check for the registry

    ureg = pint._APP_REGISTRY

    if ureg == pint._DEFAULT_REGISTRY:
        ureg = UnitRegistry()
        set_application_registry(ureg)

    Q_   = ureg.Quantity


    # length

    m   = ureg.meter
    cm  = ureg.cm
    mm  = ureg.mm
    um = ureg.micrometer


    # time

    day    = ureg.day
    hour   = ureg.hour
    minute = ureg.min
    second = ureg.second


    # frequency

    cps = ureg.count / ureg.second
    cpm = ureg.count / ureg.minute
    cph = ureg.count / ureg.hour


    # concentration

    M   = ureg.mol / ureg.L
    mM  = ureg.millimol / ureg.L
    uM = ureg.micromol / ureg.L


    # angles

    rad = ureg.rad
    deg = ureg.deg
    
    _unit_support = True
except:
    _unit_support = False



#-*- coding:utf-8 -*-
#
# plot/hive_helpers.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from matplotlib import cm
from matplotlib.colors import ColorConverter
from matplotlib.path import Path
from matplotlib.patches import PathPatch

import numpy as np

from ..lib.test_functions import nonstring_container
from .chord_diag.gradient import linear_gradient


__all__ = [
    "_get_ax_angles",
    "_get_axes_nodes",
    "_get_axes_radial_coord",
    "_get_colors",
    "_get_radial_values",
    "_get_size",
    "_plot_bezier",
    "_plot_nodes",
    "_set_names_lims",
    "RMIN"
]


RMIN = 0.1


def _get_axes_radial_coord(radial, axes, axes_bins, network):
    ''' Compute the number of axes and radial coordinates '''
    num_axes, num_radial = None, None

    # find the number of axes and radial coordinates
    if nonstring_container(radial) and len(radial) > 0:
        is_string = [isinstance(elt, str) for elt in radial]

        all_string = np.all(is_string)
        any_string = np.any(is_string)
        any_contnr = np.any([nonstring_container(elt) for elt in radial])

        if all_string:
            num_axes = num_radial = len(radial)
        elif any_string or any_contnr:
            raise ValueError("`radial` must be either a (list of) string or "
                             "a custom array of numbers.")
        else:
            num_radial = 1
    elif isinstance(radial, str):
        num_radial = 1
    else:
        raise ValueError("`radial` must be a str, list of str, or an array.")

    if nonstring_container(axes) and len(axes) > 0:
        for elt in axes:
            assert isinstance(elt, str), \
                "An `axes` list can only contain attribute names."

        assert axes_bins is None, \
            "If `axes` are given by different attributes, `axes_bins` must " \
            "be None."

        num_axes = len(axes)
    elif axes == "groups":
        assert network.structure is not None, \
            "A structure or population is required if `axes` is 'groups'."

        assert axes_bins is None, "No bins can be used if `axes` is 'groups'."

        num_axes = len(network.structure)
    elif isinstance(axes, str):
        assert axes in network.node_attributes, "Unknown `axes` attribute."

        if nonstring_container(axes_bins):
            num_axes = len(axes_bins) - 1
        elif isinstance(axes_bins, (int, np.integer)):
            num_axes = axes_bins
        else:
            raise ValueError("If `axes` is a str, then `axes_bins` must be a"
                             "valid list/number of bins.")
    else:
        raise ValueError("`axes` must be a str or list of str.")

    if num_radial > 1:
        assert num_radial == num_axes, \
            "If there is more than one radial coordinate, there must be one " \
            "per axis."

    assert num_axes >= 2, "There must be at least two axes."

    return num_axes, num_radial


def _get_axes_nodes(network, radial, axes, axes_bins, num_axes, num_radial):
    ''' Get the axes names and associated nodes and coordinates '''
    ax_names = []
    ax_nodes = []
    ax_radco = []

    if num_radial == 1:
        if axes_bins is None:
            struct = network.structure

            # each axis is a node attribute or a group
            if axes == 'groups':
                ax_names = list(network.structure)
                ax_nodes = [g.ids for g in network.structure.values()]
                ax_radco = (radial if nonstring_container(radial)
                            else network.node_attributes[radial]
                            for _ in range(num_axes))
            elif struct is not None and axes[0] in struct:
                for elt in axes:
                    assert elt in struct, \
                        "'{}' is not in the graph Structure.".format(elt)

                ax_names = list(axes)
                ax_nodes = [struct[n].ids for n in axes]
                ax_radco = (radial if nonstring_container(radial)
                            else network.node_attributes[radial]
                            for _ in range(num_axes))
            else:
                ax_names = list(axes)
                ax_nodes = [range(network.node_nb())]*num_axes
                ax_radco = [network.node_attributes[attr] for attr in axes]
        else:
            ax_radco = (radial if nonstring_container(radial)
                        else network.node_attributes[radial]
                        for _ in range(num_axes))

            axcoord = network.node_attributes[axes]

            # find node bin
            bins = axes_bins

            if isinstance(axes_bins, (int, np.integer)):
                bins = np.linspace(axcoord.min(), axcoord.max(), axes_bins + 1)

                bins[-1] += 0.01 * (bins[-1] - bins[0])

            nbin = np.digitize(axcoord, bins)

            for i in range(1, len(bins)):
                ax_nodes.append(np.where(nbin == i)[0])

            # each axis corresponds to a range of the node attribute
            for start, stop in zip(bins[:], bins[1:]):
                ax_names.append("{name}\nin [{start:.3g}, {stop:.3g}]".format(
                    name=axes, start=start, stop=stop))
    else:
        # each axis is a different radial coordinate
        ax_names = list(radial)
        ax_nodes = [range(network.node_nb())]*num_axes
        ax_radco = (network.node_attributes[attr] for attr in radial)

    return ax_names, ax_nodes, ax_radco


def _get_radial_values(ax_radco, axes_units, network):
    # r should vary between RMIN and 1 + RMIN
    radial_values = []

    if axes_units == "normed":
        for i, val in enumerate(ax_radco):
            vmin = val.min()
            vmax = val.max()
            nval = RMIN + (val - vmin) / (vmax - vmin)

            radial_values.append(nval)
    elif axes_units == "rank":
        num_nodes = network.node_nb()
        rmin = int(RMIN*num_nodes)

        for i, val in enumerate(ax_radco):
            ranks = np.argsort(np.argsort(val))
            radial_values.append((ranks + rmin) / num_nodes)
    elif axes_units == "native":
        vmin, vmax = np.inf, -np.inf

        for i, val in enumerate(ax_radco):
            vmin = min(vmin, val.min())
            vmax = max(vmax, val.max())

            # store the values as ax_radco may be a single use generator
            radial_values.append(val)

        for i, val in enumerate(radial_values):
            radial_values[i] = RMIN + (val - vmin) / (vmax - vmin)
    else:
        raise ValueError("Invalid `axes_units`: '{}'.".format(axes_units))

    return radial_values


def _smallest_angle(a1, a2):
    dtheta = np.abs(a1 - a2)

    if dtheta > np.pi:
        return 2*np.pi - dtheta

    return dtheta


def _get_ax_angles(angles, i, j, intra_connections):
    if intra_connections:
        # also works for the intra connections (i = j)
        as1 = angles[2*i]
        as2 = angles[2*i + 1]
        at1 = angles[2*j]
        at2 = angles[2*j + 1]

        if _smallest_angle(as1, at2) <= _smallest_angle(as2, at1):
            return 2*i, 2*j + 1

        return 2*i + 1, 2*j

    return i, j


def _get_size(node_size, max_nsize, ax_nodes, network):
    if node_size is None:
        max_nodes = np.max([len(nn) for nn in ax_nodes])
        return np.repeat(max(400/max_nodes, 4), network.node_nb())
    elif nonstring_container(node_size):
        assert len(node_size) == network.node_nb(), \
            "One size per node is required for array-like `node_size`."

        return np.array(node_size) / np.max(node_size) * max_nsize
    elif node_size in network.node_attributes:
        node_size = network.node_attributes[node_size]

        return node_size * (max_nsize / node_size.max())
    elif isinstance(node_size, float):
        return np.repeat(node_size, network.node_nb())

    raise ValueError("`nsize` must be float, attribute name, or array-like")


def _get_colors(axes_colors, edge_colors, angles, num_axes, intra_connections,
                network):
    ecolors = ["k"]*len(angles)
    ncolors = None

    if axes_colors is None or isinstance(axes_colors, str):
        named_cmap = "Set1" if axes_colors is None else axes_colors

        cmap = cm.get_cmap(named_cmap)

        values = list(range(num_axes))

        qualitative_cmaps = [
            "Pastel1", "Pastel2", "Accent", "Dark2", "Set1", "Set2",
            "Set3", "tab10"
        ]

        if named_cmap not in qualitative_cmaps:
            values = np.array(values) / (num_axes - 1)

        ncolors = cmap(values)
        ecolors = {}

        for i in range(num_axes):
            for j in range(num_axes):
                if i == j:
                    ecolors[(i, i)] = ncolors[i]
                else:
                    num_colors = 4 if network.is_directed() else 3

                    grad = linear_gradient(ncolors[i], ncolors[j], num_colors)

                    ecolors[(i, j)] = grad[1]
    else:
        if nonstring_container(axes_colors):
            assert len(axes_colors) == num_axes

    return ncolors, ecolors


def _set_names_lims(names, angles, max_radii, xs, ys, intra_connections,
                    show_names, axis, show_circles):
    # add names if necessary
    if show_names:
        prop = {
            "fontsize": 16*0.8,
            "ha": "center",
            "va": "center"
        }

        max_rmax = max(max_radii)

        for i, name in enumerate(names):
            angle = angles[i]
            rmax  = max_radii[i]

            if intra_connections:
                angle = 0.5*(angles[2*i] + angles[2*i + 1])
                rmax  = max_radii[2*i]

            rmax += 0.07 * (1 + name.count("\n")) * max_rmax

            x, y = rmax*np.cos(angle), rmax*np.sin(angle)

            # move to degrees
            angle *= 180 / np.pi

            if -30 <= angle <= 210:
                angle -= 90
            else:
                angle -= 270

            axis.text(x, y, name, rotation=angle, **prop)

    if not show_circles:
        for angle, rmax in zip(angles, max_radii):
            x, y = rmax*np.cos(angle), rmax*np.sin(angle)

            xs.append(x)
            ys.append(y)

        xmin = np.nanmin(xs)
        xmax = np.nanmax(xs)

        ymin = np.nanmin(ys)
        ymax = np.nanmax(ys)

        factor = 1.1

        axis.set_xlim(factor*xmin, factor*xmax)
        axis.set_ylim(factor*ymin, factor*ymax)


def _plot_nodes(nn, node_size, xx, yy, color, nborder_width, nborder_color,
                axis, zorder=3):
    if len(nn):
        ss = node_size[nn]

        axis.scatter(xx[nn], yy[nn], ss, color=color, linewidth=nborder_width,
                     edgecolors=nborder_color, zorder=zorder)


def _test_clockwise(i, j, num_axes):
    delta_max = int(0.5*num_axes)

    if num_axes == 2:
        return i != j

    for target in range(delta_max):
        for d in range(delta_max):
            if j == ((num_axes - 1 + target - d) % num_axes) and i == target:
                return True

    return False


def _plot_bezier(pstart, pstop, astart, astop, rstart, rstop, i, j, num_axes,
                 xs, ys):
    dtheta = np.abs(astart - astop)

    if dtheta > np.pi:
        dtheta = 2*np.pi - dtheta

    dtheta *= 0.3

    lstart = rstart*np.sin(dtheta)
    lstop  = rstop*np.sin(dtheta)

    dist = np.abs(i - j)

    if dist > 0.5*num_axes:
        dist -= int(0.5*num_axes)

    if _test_clockwise(i, j , num_axes):
        lstop *= -1
    elif _test_clockwise(j, i, num_axes) and dist == 1:
        lstart *= -1
    elif i > j:
        lstop *= -1
    else:
        lstart *= -1

    dp1 = np.array((lstart*np.cos(astart - 0.5*np.pi), lstart*np.sin(astart - 0.5*np.pi)))
    dp2 = np.array((lstop*np.cos(astop - 0.5*np.pi), lstop*np.sin(astop - 0.5*np.pi)))

    p1 = pstart + dp1
    p2 = pstop + dp2

    xs.append(p1[0])
    xs.append(p2[0])

    ys.append(p1[1])
    ys.append(p2[1])

    return Path([pstart, p1, p2, pstop],
                [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4])



#-*- coding:utf-8 -*-
#
# core/nx_graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Networkx subclassing """

from collections import OrderedDict, deque
from copy import deepcopy
from itertools import chain
import logging
import sys

import numpy as np
import scipy.sparse as ssp

import nngt
from nngt.lib import InvalidArgument, BWEIGHT, nonstring_container, is_integer
from nngt.lib.connect_tools import (_cleanup_edges, _set_dist_new_edges,
                                    _set_default_edge_attributes)
from nngt.lib.graph_helpers import (_get_dtype, _get_nx_weights,
                                    _post_del_update)
from nngt.lib.converters import _np_dtype, _to_np_array
from nngt.lib.logger import _log_message
from .graph_interface import GraphInterface, BaseProperty


logger = logging.getLogger(__name__)


# ---------- #
# Properties #
# ---------- #

class _NxNProperty(BaseProperty):

    '''
    Class for generic interactions with nodes properties (networkx)
    '''

    def __getitem__(self, name):
        g = self.parent()._graph

        lst = [g.nodes[i][name] for i in range(g.number_of_nodes())]

        dtype = _np_dtype(super(_NxNProperty, self).__getitem__(name))

        return _to_np_array(lst, dtype=dtype)

    def __setitem__(self, name, value):
        g    = self.parent()._graph
        size = g.number_of_nodes()

        if name in self:
            if len(value) == size:
                for i in range(size):
                    g.nodes[i][name] = value[i]
            else:
                raise ValueError("A list or a np.array with one entry per "
                                 "node in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

    def new_attribute(self, name, value_type, values=None, val=None):
        g = self.parent()._graph

        if val is None:
            if value_type == "int":
                val = int(0)
            elif value_type == "double":
                val = np.NaN
            elif value_type == "string":
                val = ""
            else:
                val = None
                value_type = "object"

        if values is None:
            values = [deepcopy(val)
                      for _ in range(g.number_of_nodes())]

        # store name and value type in the dict
        super(_NxNProperty, self).__setitem__(name, value_type)
        # store the real values in the attribute
        self[name] = values
        self._num_values_set[name] = len(values)

    def set_attribute(self, name, values, nodes=None):
        '''
        Set the node attribute.

        Parameters
        ----------
        name : str
            Name of the node attribute.
        values : array, size N
            Values that should be set.
        nodes : array-like, optional (default: all nodes)
            Nodes for which the value of the property should be set. If `nodes`
            is not None, it must be an array of size N.
        '''
        g = self.parent()._graph

        num_nodes = g.number_of_nodes()
        num_n = len(nodes) if nodes is not None else num_nodes

        if num_n == num_nodes:
            self[name] = values
        else:
            if num_n != len(values):
                raise ValueError("`nodes` and `values` must have the same "
                                 "size; got respectively " + str(num_n) + \
                                 " and " + str(len(values)) + " entries.")
            else:
                for n, val in zip(nodes, values):
                    g.nodes[n][name] = val
        self._num_values_set[name] = num_nodes


class _NxEProperty(BaseProperty):

    ''' Class for generic interactions with edge properties (networkx)  '''

    def __getitem__(self, name):
        g     = self.parent()._graph
        edges = None

        if isinstance(name, slice):
            edges = self.parent().edges_array[name]
        elif nonstring_container(name):
            if len(name) == 0:
                return []

            if nonstring_container(name[0]):
                edges = name
            else:
                if len(name) != 2:
                    raise InvalidArgument(
                        "key for edge attribute must be one of the following: "
                        "slice, list of edges, edges or attribute name.")
                return g[name[0]][name[1]]

        if isinstance(name, str):
            dtype = _np_dtype(super(_NxEProperty, self).__getitem__(name))
            eprop = np.empty(g.number_of_edges(), dtype=dtype)

            edges = list(g.edges(data=name))

            if len(edges):
                eids  = np.asarray(list(g.edges(data="eid")))[:, 2]

                for i, eid in enumerate(np.argsort(eids)):
                    eprop[i] = edges[eid][2]

            return eprop

        eprop = {k: [] for k in self.keys()}

        for edge in edges:
            data = g.get_edge_data(edge[0], edge[1])

            if data is None:
                raise ValueError("Edge {} does not exist.".format(edge))

            for k, v in data.items():
                if k != "eid":
                    eprop[k].append(v)

        dtype = None

        for k, v in eprop.items():
            dtype    = _np_dtype(super(_NxEProperty, self).__getitem__(k))
            eprop[k] = _to_np_array(v, dtype)

        return eprop

    def __setitem__(self, name, value):
        g = self.parent()._graph

        if name in self:
            size = g.number_of_edges()

            if len(value) == size:
                edges = np.asarray(list(g.edges(data="eid")))

                if size:
                    order = np.argsort(edges[:, 2])

                    for i, idx in enumerate(order):
                        s, t, _ = edges[idx]
                        g.edges[s, t][name] = value[i]
            else:
                raise ValueError(
                    "A list or a np.array with one entry per edge in the "
                    "graph is required. For attribute "
                    "'{}', got {} entries vs {} edges.".format(
                        name, len(value), size))
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

    def new_attribute(self, name, value_type, values=None, val=None):
        g = self.parent()._graph

        if val is None:
            if value_type == "int":
                val = int(0)
            elif value_type == "double":
                val = np.NaN
            elif value_type == "string":
                val = ""
            else:
                val = None
                value_type = "object"

        if values is None:
            values = [deepcopy(val)
                      for _ in range(g.number_of_edges())]

        # store name and value type in the dict
        super(_NxEProperty, self).__setitem__(name, value_type)

        # store the real values in the attribute
        self[name] = values
        self._num_values_set[name] = len(values)

    def set_attribute(self, name, values, edges=None, last_edges=False):
        '''
        Set the edge property.

        Parameters
        ----------
        name : str
            Name of the edge property.
        values : array
            Values that should be set.
        edges : array-like, optional (default: None)
            Edges for which the value of the property should be set. If `edges`
            is not None, it must be an array of shape `(len(values), 2)`.
        '''
        graph = self.parent()

        g = graph._graph

        num_edges = g.number_of_edges()
        num_e = len(edges) if edges is not None else num_edges

        if num_e != len(values):
            raise ValueError("`edges` and `values` must have the same "
                             "size; got respectively " + str(num_e) + \
                             " and " + str(len(values)) + " entries.")

        if edges is None:
            self[name] = values
        else:
            order = range(num_e)

            if not last_edges:
                get_eid = graph.edge_id

                eids  = [get_eid(e) for e in edges]
                order = np.argsort(np.argsort(eids))

            for i, e in zip(order, edges):
                edict = g[e[0]][e[1]]

                edict[name]  = values[i]

                g.add_edge(e[0], e[1], **edict)

        if num_e:
            self._num_values_set[name] = num_edges


# ----- #
# Graph #
# ----- #

class _NxGraph(GraphInterface):

    '''
    Subclass of networkx Graph
    '''

    _nattr_class = _NxNProperty
    _eattr_class = _NxEProperty

    #-------------------------------------------------------------------------#
    # Class properties

    di_value = {"string": "", "double": 0., "int": int(0)}

    #-------------------------------------------------------------------------#
    # Constructor and instance properties

    def __init__(self, nodes=0, copy_graph=None, directed=True, weighted=False,
                 **kwargs):
        self._nattr = _NxNProperty(self)
        self._eattr = _NxEProperty(self)

        self._max_eid = 0

        g = copy_graph.graph if copy_graph is not None else None

        if g is not None:
            if not directed and g.is_directed():
                g = g.to_undirected()
            elif directed and not g.is_directed():
                g = g.to_directed()

            self._from_library_graph(g, copy=True)

            self._max_eid = copy_graph._max_eid
        else:
            nx = nngt._config["library"]

            self._graph = nx.DiGraph() if directed else nx.Graph()

            if nodes:
                self._graph.add_nodes_from(range(nodes))

    #-------------------------------------------------------------------------#
    # Graph manipulation

    def edge_id(self, edge):
        '''
        Return the ID a given edge or a list of edges in the graph.
        Raises an error if the edge is not in the graph or if one of the
        vertices in the edge is nonexistent.

        Parameters
        ----------
        edge : 2-tuple or array of edges
            Edge descriptor (source, target).

        Returns
        -------
        index : int or array of ints
            Index of the given `edge`.
        '''
        g = self._graph

        if is_integer(edge[0]):
            return g[edge[0]][edge[1]]["eid"]
        elif nonstring_container(edge[0]):
            return [g[e[0]][e[1]]["eid"] for e in edge]

        raise AttributeError("`edge` must be either a 2-tuple of ints or "
                             "an array of 2-tuples of ints.")

    def has_edge(self, edge):
        '''
        Whether `edge` is present in the graph.

        .. versionadded:: 2.0
        '''
        return self._graph.has_edge(*edge)

    @property
    def edges_array(self):
        '''
        Edges of the graph, sorted by order of creation, as an array of
        2-tuple.
        '''
        g     = self._graph
        edges = np.full((self._max_eid, 2), -1)

        # fast iteration using list comprehension
        # could also be done with deque and map (deque forces lazy map to run)
        # deque(map(lambda x: _gen_edges(edges, x), g.edges(data="eid")))
        [_gen_edges(edges, x) for x in g.edges(data="eid")]

        if self._max_eid > g.number_of_edges():
            return edges[edges[:, 0] > -1]

        return edges

    def _get_edges(self, source_node=None, target_node=None):
        '''
        Called by Graph.get_edges if source_node and target_node are not both
        integers.
        '''
        nx = nngt._config["library"]

        g = self._graph

        target_node = \
            [target_node] if is_integer(target_node) else target_node

        if source_node is not None:
            source_node = \
                [source_node] if is_integer(source_node) else source_node

            if target_node is None:

                if g.is_directed():
                    return list(g.out_edges(source_node))

                return [
                    e if e[0] <= e[1] else e[::-1] for e in g.edges(source_node)
                ]

            res_iter = nx.edge_boundary(g, source_node, target_node)

            if g.is_directed():
                return list(res_iter)

            return [e if e[0] <= e[1] else e[::-1] for e in res_iter]

        if target_node is None:
            # return all edges
            return list(g.edges)

        if g.is_directed():
            return list(g.in_edges(target_node))

        return [
            e if e[0] <= e[1] else e[::-1] for e in g.edges(target_node)
        ]

    def new_node(self, n=1, neuron_type=1, attributes=None, value_types=None,
                 positions=None, groups=None):
        '''
        Adding a node to the graph, with optional properties.

        Parameters
        ----------
        n : int, optional (default: 1)
            Number of nodes to add.
        neuron_type : int, optional (default: 1)
            Type of neuron (1 for excitatory, -1 for inhibitory)
        attributes : dict, optional (default: None)
            Dictionary containing the attributes of the nodes.
        value_types : dict, optional (default: None)
            Dict of the `attributes` types, necessary only if the `attributes`
            do not exist yet.
        positions : array of shape (n, 2), optional (default: None)
            Positions of the neurons. Valid only for
            :class:`~nngt.SpatialGraph` or :class:`~nngt.SpatialNetwork`.
        groups : str, int, or list, optional (default: None)
            :class:`~nngt.core.NeuralGroup` to which the neurons belong. Valid
            only for :class:`~nngt.Network` or :class:`~nngt.SpatialNetwork`.

        Returns
        -------
        The node or a list of the nodes created.
        '''
        g = self._graph
        num_nodes = g.number_of_nodes()

        new_nodes = list(range(num_nodes, num_nodes + n))

        for v in new_nodes:
            g.add_node(v)

        attributes = {} if attributes is None else deepcopy(attributes)

        if attributes:
            for k, v in attributes.items():
                if k not in self._nattr:
                    self._nattr.new_attribute(k, value_types[k], val=v)
                else:
                    v = v if nonstring_container(v) else [v]
                    self._nattr.set_attribute(k, v, nodes=new_nodes)

        # set default values for all attributes that were not set
        for k in self.node_attributes:
            if k not in attributes:
                dtype = self.get_attribute_type(k)
                filler = [None for _ in new_nodes]

                # change for strings, doubles and ints
                if dtype == "string":
                    filler = ["" for _ in new_nodes]
                elif dtype == "double":
                    filler = [np.NaN for _ in new_nodes]
                elif dtype == "int":
                    filler = [0 for _ in new_nodes]

                self._nattr.set_attribute(k, filler, nodes=new_nodes)

        if self.is_spatial():
            old_pos      = self._pos
            self._pos    = np.full((self.node_nb(), 2), np.NaN)
            num_existing = len(old_pos) if old_pos is not None else 0

            if num_existing != 0:
                self._pos[:num_existing] = old_pos

        if positions is not None and len(positions):
            assert self.is_spatial(), \
                "`positions` argument requires a SpatialGraph/SpatialNetwork."
            self._pos[new_nodes] = positions

        if groups is not None:
            assert self.is_network(), \
                "`positions` argument requires a Network/SpatialNetwork."
            if nonstring_container(groups):
                assert len(groups) == n, "One group per neuron required."
                for g, node in zip(groups, new_nodes):
                    self.population.add_to_group(g, node)
            else:
                self.population.add_to_group(groups, new_nodes)

        if len(new_nodes) == 1:
            return new_nodes[0]

        return new_nodes

    def delete_nodes(self, nodes):
        '''
        Remove nodes (and associated edges) from the graph.
        '''
        g = self._graph

        if nonstring_container(nodes):
            for n in nodes:
                g.remove_node(n)
        else:
            g.remove_node(nodes)

        # relabel nodes from zero
        nx = nngt._config["library"]

        nx.relabel_nodes(g, {n: i for i, n in enumerate(g.nodes)}, copy=False)

        # update attributes
        for key in self._nattr:
            self._nattr._num_values_set[key] = self.node_nb()

        for key in self._eattr:
            self._eattr._num_values_set[key] = self.edge_nb()

        # check spatial and structure properties
        _post_del_update(self, nodes)

    def new_edge(self, source, target, attributes=None, ignore=False,
                 self_loop=False):
        '''
        Adding a connection to the graph, with optional properties.

        .. versionchanged :: 2.0
            Added `self_loop` argument to enable adding self-loops.

        Parameters
        ----------
        source : :class:`int/node`
            Source node.
        target : :class:`int/node`
            Target node.
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": 1.}``, the unit weight for the
            connection (synaptic strength in NEST).
        ignore : bool, optional (default: False)
            If set to True, ignore attempts to add an existing edge and accept
            self-loops; otherwise an error is raised.
        self_loop : bool, optional (default: False)
            Whether to allow self-loops or not.

        Returns
        -------
        The new connection or None if nothing was added.
        '''
        g = self._graph

        attributes = {} if attributes is None else deepcopy(attributes)

        # check that nodes exist
        num_nodes = g.number_of_nodes()

        if source >= num_nodes or target >= num_nodes:
            raise InvalidArgument("`source` or `target` does not exist.")

        # set default values for attributes that were not passed
        _set_default_edge_attributes(self, attributes, num_edges=1)

        if g.has_edge(source, target):
            if not ignore:
                raise InvalidArgument("Trying to add existing edge.")

            _log_message(logger, "INFO",
                         "Existing edge {} ignored.".format((source, target)))
        else:
            if source == target:
                if not ignore and not self_loop:
                    raise InvalidArgument("Trying to add a self-loop.")
                elif ignore:
                    _log_message(logger, "INFO",
                                 "Self-loop on {} ignored.".format(source))

                    return None

            for attr in attributes:
                if "_corr" in attr:
                    raise NotImplementedError("Correlated attributes are not "
                                              "available with networkx.")

            if self.is_weighted() and "weight" not in attributes:
                attributes["weight"] = 1.

            # check distance
            _set_dist_new_edges(attributes, self, [(source, target)])

            g.add_edge(source, target)

            g[source][target]["eid"] = self._max_eid

            self._max_eid += 1

            # call parent function to set the attributes
            self._attr_new_edges([(source, target)], attributes=attributes)

        return (source, target)

    def new_edges(self, edge_list, attributes=None, check_duplicates=False,
                  check_self_loops=True, check_existing=True,
                  ignore_invalid=False):
        '''
        Add a list of edges to the graph.

        .. versionchanged:: 2.0
            Can perform all possible checks before adding new edges via the
            ``check_duplicates`` ``check_self_loops``, and ``check_existing``
            arguments.

        Parameters
        ----------
        edge_list : list of 2-tuples or np.array of shape (edge_nb, 2)
            List of the edges that should be added as tuples (source, target)
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": ones}``, where ``ones`` is an
            array the same length as the `edge_list` containing a unit weight
            for each connection (synaptic strength in NEST).
        check_duplicates : bool, optional (default: False)
            Check for duplicate edges within `edge_list`.
        check_self_loops : bool, optional (default: True)
            Check for self-loops.
        check_existing : bool, optional (default: True)
            Check whether some of the edges in `edge_list` already exist in the
            graph or exist multiple times in `edge_list` (also performs
            `check_duplicates`).
        ignore_invalid : bool, optional (default: False)
            Ignore invalid edges: they are not added to the graph and are
            silently dropped. Unless this is set to true, an error is raised
            whenever one of the three checks fails.

        .. warning::

            Setting `check_existing` to False will lead to undefined behavior
            if existing edges are provided! Only use it (for speedup) if you
            are sure that you are indeed only adding new edges.

        Returns
        -------
        Returns new edges only.
        '''
        g = self._graph

        attributes = {} if attributes is None else deepcopy(attributes)
        num_edges  = len(edge_list)

        # check that all nodes exist
        if num_edges:
            if np.max(edge_list) >= g.number_of_nodes():
                raise InvalidArgument("Some nodes do no exist.")

            for attr in attributes:
                if "_corr" in attr:
                    raise NotImplementedError("Correlated attributes are not "
                                              "available with networkx.")

            # set default values for attributes that were not passed
            _set_default_edge_attributes(self, attributes, num_edges)

            # check edges
            new_attr = None

            if check_duplicates or check_self_loops or check_existing:
                edge_list, new_attr = _cleanup_edges(
                    self, edge_list, attributes, check_duplicates,
                    check_self_loops, check_existing, ignore_invalid)
            else:
                new_attr = attributes

            # create the edges
            initial_eid = self._max_eid

            num_added = len(edge_list)

            if num_added:
                arr_edges = np.zeros((num_added, 3), dtype=int)

                arr_edges[:, :2] = edge_list
                arr_edges[:, 2]  = np.arange(
                    initial_eid, initial_eid + num_added)

                # create the edges with an eid attribute
                g.add_weighted_edges_from(arr_edges, weight="eid")

                # check distance
                _set_dist_new_edges(new_attr, self, edge_list)

                # call parent function to set the attributes
                self._attr_new_edges(edge_list, attributes=new_attr)

                self._max_eid += num_added

        return edge_list

    def delete_edges(self, edges):
        ''' Remove a list of edges '''
        if nonstring_container(edges[0]):
            self._graph.remove_edges_from(edges)
        else:
            self._graph.remove_edge(*edges)

        for key in self._eattr:
            self._eattr._num_values_set[key] = self.edge_nb()

    def clear_all_edges(self):
        ''' Remove all edges from the graph '''
        g = self._graph
        g.remove_edges_from(tuple(g.edges()))
        self._eattr.clear()

    #-------------------------------------------------------------------------#
    # Getters

    def node_nb(self):
        ''' Number of nodes in the graph '''
        return self._graph.number_of_nodes()

    def edge_nb(self):
        ''' Number of edges in the graph '''
        return self._graph.number_of_edges()

    def get_degrees(self, mode="total", nodes=None, weights=None):
        g = self._graph
        w = _get_nx_weights(self, weights)

        nodes  = range(g.number_of_nodes()) if nodes is None else nodes
        dtype  = int if weights in {False, None} else float
        di_deg = None

        if mode == 'total' or not self._graph.is_directed():
            di_deg = g.degree(nodes, weight=w)
        elif mode == 'in':
            di_deg = g.in_degree(nodes, weight=w)
        elif mode == 'out':
            di_deg = g.out_degree(nodes, weight=w)
        else:
            raise ValueError("Unknown `mode` '{}'".format(mode))

        if nonstring_container(nodes):
            return np.array([di_deg[i] for i in nodes], dtype=dtype)

        return di_deg

    def is_connected(self, mode="strong"):
        '''
        Return whether the graph is connected.

        Parameters
        ----------
        mode : str, optional (default: "strong")
            Whether to test connectedness with directed ("strong") or
            undirected ("weak") connections.
        '''
        g = self._graph

        if g.is_directed() and mode == "weak":
            g = g.to_undirected(as_view=True)

        try:
            import networkx as nx
            nx.diameter(g)
            return True
        except nx.exception.NetworkXError:
            return False

    def neighbours(self, node, mode="all"):
        '''
        Return the neighbours of `node`.

        Parameters
        ----------
        node : int
            Index of the node of interest.
        mode : string, optional (default: "all")
            Type of neighbours that will be returned: "all" returns all the
            neighbours regardless of directionality, "in" returns the
            in-neighbours (also called predecessors) and "out" retruns the
            out-neighbours (or successors).

        Returns
        -------
        neighbours : set
            The neighbours of `node`.
        '''
        g = self._graph

        # special case for undirected
        if not g.is_directed():
            return set(g.neighbors(node))

        if mode == "all":
            # for directed graphs, neighbors ~ successors
            return set(g.successors(node)).union(g.predecessors(node))
        elif mode == "in":
            return set(g.predecessors(node))
        elif mode == "out":
            return set(g.successors(node))

        raise ArgumentError('Invalid `mode` argument {}; possible values are '
                            '"all", "out" or "in".'.format(mode))

    def _from_library_graph(self, graph, copy=True):
        ''' Initialize `self._graph` from existing library object. '''
        import networkx as nx

        nodes = {n: i for i, n in enumerate(graph)}

        num_nodes = graph.number_of_nodes()
        num_edges = graph.number_of_edges()

        # check if nodes start from 0 and are continuous
        if set(nodes.keys()) != set(range(num_nodes)):
            # forced copy to restore nodes to [0, num_nodes[
            g = None

            if graph.is_directed():
                g = nx.DiGraph()
            else:
                g = nx.Graph()

            # add nodes
            for i, (n, attr) in enumerate(graph.nodes(data=True)):
                attr["id"] = n
                g.add_node(i, **attr)

            # add edges
            [g.add_edge(nodes[u], nodes[v], **attr)
             for u, v, attr in graph.edges(data=True)]

            # make edges ids
            def set_eid(e, eid):
                g.edges[e]["eid"] = eid

            [set_eid(e, i) for i, e in enumerate(g.edges)]

            self._max_eid = num_edges

            self._graph = graph = g
        else:
            # all good
            self._graph = graph.copy() if copy else graph

        # get attributes names and "types" and initialize them
        if num_nodes:
            for key, val in graph.nodes[0].items():
                super(type(self._nattr), self._nattr).__setitem__(
                    key, _get_dtype(val))

        if num_edges:
            e0 = next(iter(graph.edges))

            for key, val in graph.edges[e0].items():
                if key != "eid":
                    super(type(self._eattr), self._eattr).__setitem__(
                        key, _get_dtype(val))


# tool function to generate the edges_array

def _gen_edges(array, edata):
    source, target, eid = edata
    array[eid] = (source, target)



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_rewire.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Test the rewire methods of the :mod:`~nngt.generation` module.
"""

import os

import numpy as np
import pytest

import nngt
import nngt.analysis as na
import nngt.generation as ng


if os.environ.get("MPI"):
    nngt.set_config("mpi", True)


@pytest.mark.mpi_skip
def test_random_rewire():
    ''' Check random rewire '''
    num_nodes = 10
    coord_nb  = 2
    recip     = 0.7
    shortcut  = 0.2

    g = ng.newman_watts(coord_nb, shortcut, reciprocity_circular=recip,
                        nodes=num_nodes)

    # get graph properties
    final_recip = na.reciprocity(g)
    num_edges   = g.edge_nb()

    # make some node and edge attributes
    g.new_node_attribute("random_int", "int",
                         values=[2, 5, 33, 6, 4, 1, 98, 45, 30, 10])
    g.new_node_attribute("attr2", "float",
                         values=nngt._rng.uniform(size=num_nodes))

    ww = np.arange(1, num_edges + 1, dtype=float)
    g.set_weights(ww)

    g.new_edge_attribute("my-edge-attr", "int", values=-ww[::-1].astype(int))

    # completely random rewiring
    r1 = ng.random_rewire(g)

    assert r1.node_nb() == num_nodes
    assert r1.edge_nb() == num_edges
    assert g.node_attributes == r1.node_attributes
    assert g.edge_attributes == r1.edge_attributes
    assert not np.array_equal(g.node_attributes["random_int"],
                              r1.node_attributes["random_int"])
    assert not np.array_equal(g.node_attributes["attr2"],
                              r1.node_attributes["attr2"])
    assert not np.array_equal(g.edge_attributes["weight"],
                              r1.edge_attributes["weight"])
    assert not np.array_equal(g.edge_attributes["my-edge-attr"],
                              r1.edge_attributes["my-edge-attr"])

    # keep degrees
    for deg_type in ["in-degree", "out-degree", "total-degree"]:
        degrees  = g.get_degrees(deg_type)

        edge_constraint = "together"

        if deg_type == "in-degree":
            edge_constraint = "preserve_in"
        elif deg_type == "out-degree":
            edge_constraint = "preserve_out"

        r2 = ng.random_rewire(g, constraints=deg_type,
                              node_attr_constraints="preserve",
                              edge_attr_constraints=edge_constraint)

        # check basics
        assert r2.node_nb() == num_nodes
        assert r2.edge_nb() == num_edges

        assert np.array_equal(r2.get_degrees(deg_type), degrees)

        # check node attributes
        assert g.node_attributes == r2.node_attributes
        assert np.array_equal(g.node_attributes["random_int"],
                              r2.node_attributes["random_int"])
        assert np.array_equal(g.node_attributes["attr2"],
                              r2.node_attributes["attr2"])

        # check edge attributes
        if deg_type == "total-degree":
            # check that attributes were moved together
            weights  = r2.get_weights()
            my_eattr = r2.edge_attributes["my-edge-attr"]

            assert np.array_equal(my_eattr,
                                  (weights - (num_edges + 1)).astype(int))
        else:
            node_type = ("source_node" if edge_constraint == "preserve_out"
                         else "target_node")

            for i in range(num_nodes):
                kwarg = {node_type: i}

                old_edges = g.get_edges(**kwarg)
                new_edges = r2.get_edges(**kwarg)

                for name in g.edge_attributes:
                    old_attr = g.get_edge_attributes(old_edges, name)
                    new_attr = r2.get_edge_attributes(new_edges, name)

                    assert set(old_attr) == set(new_attr)


@pytest.mark.mpi_skip
def test_complete_lattice_rewire():
    ''' Check lattice rewiring method. '''
    num_nodes = 10
    degree = 4

    g = ng.fixed_degree(degree, "total", nodes=num_nodes)

    # node attributes
    g.new_node_attribute("random_int", "int",
                         values=[2, 5, 33, 6, 4, 1, 98, 45, 30, 10])
    g.new_node_attribute("attr2", "float",
                         values=nngt._rng.uniform(size=num_nodes))

    # edge attributes
    ww = nngt._rng.uniform(1, 5, size=g.edge_nb())
    g.set_weights(ww)
    g.new_edge_attribute("my-edge-attr", "int", values=-ww.astype(int))

    # rewire
    l1 = ng.lattice_rewire(g, weight="weight",
                           node_attr_constraints="preserve")

    assert g.node_nb() == l1.node_nb()
    assert g.edge_nb() == l1.edge_nb()

    # check node attributes
    assert g.node_attributes == l1.node_attributes
    assert np.array_equal(g.node_attributes["random_int"],
                          l1.node_attributes["random_int"])
    assert np.array_equal(g.node_attributes["attr2"],
                          l1.node_attributes["attr2"])

    # check the order of the first weights
    srt = np.sort(l1.get_weights())[::-1]
    oth = -srt.astype(int)

    all_other = True

    for i in range(num_nodes - 1):
        e = (i,  i + 1)
        assert l1.get_edge_attributes(e, name='weight') == srt[2*i]
        all_other *= l1.get_edge_attributes(e, name='my-edge-attr') == oth[2*i]
        
        e = (i + 1, i)
        assert l1.get_edge_attributes(e, name='weight') == srt[2*i + 1]
        all_other *= (l1.get_edge_attributes(e, name='my-edge-attr')
                      == oth[2*i + 1])

    assert not all_other


@pytest.mark.mpi_skip
def test_incomplete_lattice_rewire():
    ''' Lattice rewire when node degrees are odd. '''
    num_nodes = 10
    degree = 7

    g = ng.fixed_degree(degree, "total", nodes=num_nodes)

    # node attributes
    g.new_node_attribute("random_int", "int",
                         values=[2, 5, 33, 6, 4, 1, 98, 45, 30, 10])
    g.new_node_attribute("attr2", "float",
                         values=nngt._rng.uniform(size=num_nodes))

    # edge attributes
    ww = nngt._rng.uniform(1, 5, size=g.edge_nb())
    g.set_weights(ww)
    g.new_edge_attribute("my-edge-attr", "int", values=-ww.astype(int))

    # rewire
    l2 = ng.lattice_rewire(g, weight="weight", distance_sort="linear",
                           edge_attr_constraints="together")

    assert g.node_nb() == l2.node_nb()
    assert g.edge_nb() == l2.edge_nb()

    # check node attributes (may fail every million trials or so)
    assert g.node_attributes == l2.node_attributes
    assert not np.array_equal(g.node_attributes["random_int"],
                              l2.node_attributes["random_int"])
    assert not np.array_equal(g.node_attributes["attr2"],
                              l2.node_attributes["attr2"])

    # check edge attributes
    mea = l2.edge_attributes["my-edge-attr"]
    assert np.array_equal(mea, -l2.get_weights().astype(int))

    # check the order of the first weights
    srt = np.sort(ww)
    oth = -srt.astype(int)

    for i in range(num_nodes - 1):
        e = (i,  i + 1)
        assert l2.get_edge_attributes(e, name='weight') == srt[2*i]
        assert l2.get_edge_attributes(e, name='my-edge-attr') == oth[2*i]
        
        e = (i + 1, i)
        assert l2.get_edge_attributes(e, name='weight') == srt[2*i + 1]
        assert l2.get_edge_attributes(e, name='my-edge-attr') == oth[2*i + 1]


@pytest.mark.mpi_skip
def test_reciprocity_lattice_rewire():
    ''' Lattice rewire with variable reciprocity. '''
    num_nodes = 1000
    degree = 7

    g = ng.fixed_degree(degree, "total", nodes=num_nodes)

    # edge attributes
    ww = nngt._rng.uniform(1, 5, size=g.edge_nb())
    g.set_weights(ww)
    g.new_edge_attribute("my-edge-attr", "int", values=-ww.astype(int))

    # rewire
    reciprocity = 0.7

    l = ng.lattice_rewire(g, weight="weight", target_reciprocity=reciprocity,
                          edge_attr_constraints="together")

    assert np.isclose(na.reciprocity(l), reciprocity)

    srt = np.sort(ww)[::-1]
    oth = -srt.astype(int)

    for i in range(num_nodes - 1):
        e = (i,  i + 1)
        assert l.get_edge_attributes(e, name='weight') == srt[2*i]
        assert l.get_edge_attributes(e, name='my-edge-attr') == oth[2*i]
        
        e = (i + 1, i)
        assert l.get_edge_attributes(e, name='weight') == srt[2*i + 1]
        assert l.get_edge_attributes(e, name='my-edge-attr') == oth[2*i + 1]


if __name__ == "__main__":
    if not nngt.get_config("mpi"):
        test_random_rewire()
        test_complete_lattice_rewire()
        test_incomplete_lattice_rewire()
        test_reciprocity_lattice_rewire()



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the PyNCulture project, which aims at providing tools to
# easily generate complex neuronal cultures.
# Copyright (C) 2017 SENeC Initiative
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Tools for PyNCulture """

try:
    from collections.abc import Container as _container
except:
    from collections import Container as _container

import numpy as np

from . import _shapely_support


def indexable(obj):
    '''
    Returns true for any iterable which is not a string or byte sequence.
    '''
    if hasattr(obj, "__getitem__"):
        return True
    return False


def pop_largest(shapes):
    '''
    Returns the largest shape, removing it from the list.
    If `shapes` is a :class:`shapely.geometry.MultiPolygon`, returns the
    largest :class:`shapely.geometry.Polygon` without modifying the object.

    .. versionadded:: 0.3

    Parameters
    ----------
    shapes : list of :class:`Shape` objects or MultiPolygon.
    '''
    MultiPolygon = None
    try:
        from shapely.geometry import MultiPolygon
    except ImportError:
        pass

    max_area = -np.inf
    max_idx  = -1

    for i, s in enumerate(shapes):
        if s.area > max_area:
            max_area = s.area
            max_idx  = i

    if shapes.__class__ == MultiPolygon:
        return shapes[max_idx]

    return shapes.pop(max_idx)


def _insert_area(container, area_name, shape, height, properties):
    '''
    Insert the area into the container, potentially restructuring the existing
    areas.
    In particular, if the shape is composed of multiple polygons, it will
    be inserted as "area_name_X", with X ranging from 0 to N-1, N being the
    number of polygons.

    If `area_name` already exists in `container`, it will be overriden,
    potentially even deleted in favor of numbered subareas if it is
    replaced by a set of polygons.
    The only exception to that rule is the "default_area", which is never
    deleted but replaced by the largest polygon, while the smaller ones are
    numbered from 1.
    '''
    # import
    from .shape import Area
    from shapely.geometry import MultiPolygon
    # check for multiple polygons
    if shape.__class__ == MultiPolygon:
        # behavior differs for default_area (never deleted) and other areas
        if area_name == "default_area":
            largest = pop_largest(shape)
            count   = len(container.default_areas)
            for p in shape:
                new_name = area_name
                if p != largest:
                    new_name = area_name + '_' + str(count)
                    count += 1
                container._areas[new_name] = Area.from_shape(
                    p, height=height, name=new_name, properties=properties)
        else:
            for i, p in enumerate(shape):
                new_name = area_name + '_' + str(i)
                container._areas[new_name] = Area.from_shape(
                    p, height=height, name=new_name, properties=properties)
            if area_name in container.areas:
                del container._areas[area_name]
    else:
        container._areas[area_name] = Area.from_shape(
                shape, height=height, name=area_name, properties=properties)
            

def _backup_contains(x, y, shape):
    try:
        x = np.array(x)
        y = np.array(y)
    except:
        pass
    if shape.geom_type == "Disk":
        x0, y0           = shape.centroid
        xmin, _, xmax, _ = shape.bounds
        radius = 0.5*(xmax - xmin)
        return np.less_equal(np.linalg.norm([x - x0, y - y0], axis=0), radius)
    elif shape.geom_type == "Ellipse":
        xmin, ymin, xmax, ymax = shape.bounds
        a      = 0.5*(xmax - xmin)
        b      = 0.5*(ymax - ymin)
        x0, y0 = shape.centroid
        return np.less_equal(np.square(x-x0) / a + np.square(y-y0) / b, 1.)
    elif shape.geom_type == "Rectangle":
        xmin, ymin, xmax, ymax = shape.bounds
        contained  = np.less_equal(x, xmax)
        contained *= np.greater_equal(x, xmin)
        contained *= np.less_equal(y, ymax)
        contained *= np.greater_equal(y, ymin)
        return contained
    else:
        raise TypeError("Invalid Shape type: {}.".format(shape.geom_type))
        



"""
Create linear color gradients
"""

from matplotlib.colors import ColorConverter, LinearSegmentedColormap
from scipy.ndimage import gaussian_filter

import numpy as np


def linear_gradient(cstart, cend, n=10):
    '''
    Return a gradient list of `n` colors going from `cstart` to `cend`.
    '''
    s = np.array(ColorConverter.to_rgb(cstart))
    f = np.array(ColorConverter.to_rgb(cend))

    rgb_list = [s + (t / (n - 1))*(f - s) for t in range(n)]

    return rgb_list


def gradient(start, end, min_angle, color1, color2, meshgrid, mask, ax,
             alpha):
    '''
    Create a linear gradient from `start` to `end`, which is translationally
    invarient in the orthogonal direction.
    The gradient is then cliped by the mask.
    '''
    xs, ys = start
    xe, ye = end

    X, Y = meshgrid

    # get the distance to each point
    d2start = (X - xs)*(X - xs) + (Y - ys)*(Y - ys)
    d2end   = (X - xe)*(X - xe) + (Y - ye)*(Y - ye)

    dmax = (xs - xe)*(xs - xe) + (ys - ye)*(ys - ye)

    # blur
    smin = 0.015*len(X)
    smax = max(smin, 0.1*len(X)*min(min_angle/120, 1))

    sigma = np.clip(dmax*len(X), smin, smax)

    Z = gaussian_filter((d2end < d2start).astype(float), sigma=sigma)

    # generate the colormap
    n_bin = 100

    color_list = linear_gradient(color1, color2, n_bin)

    cmap = LinearSegmentedColormap.from_list("gradient", color_list, N=n_bin)

    im = ax.imshow(Z, interpolation='bilinear', cmap=cmap,
                   origin='lower', extent=[-1, 1, -1, 1], alpha=alpha)

    im.set_clip_path(mask)



#-*- coding:utf-8 -*-
#
# core/networks.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Network and SpatialNetwork classes for neuroscience integration """

import numpy as np

import nngt
from nngt.lib import (InvalidArgument, nonstring_container, default_neuron,
                      default_synapse)
from .graph import Graph
from .spatial_graph import SpatialGraph


# ------- #
# Network #
# ------- #

class Network(Graph):

    """
    The detailed class that inherits from :class:`~nngt.Graph` and implements
    additional properties to describe various biological functions
    and interact with the NEST simulator.
    """

    #-------------------------------------------------------------------------#
    # Class attributes and methods

    __num_networks = 0
    __max_id       = 0

    @classmethod
    def num_networks(cls):
        ''' Returns the number of alive instances. '''
        return cls.__num_networks

    @classmethod
    def from_gids(cls, gids, get_connections=True, get_params=False,
                  neuron_model=default_neuron, neuron_param=None,
                  syn_model=default_synapse, syn_param=None, **kwargs):
        '''
        Generate a network from gids.

        Warning
        -------
        Unless `get_connections` and `get_params` is True, or if your
        population is homogeneous and you provide the required information, the
        information contained by the network and its `population` attribute
        will be erroneous!
        To prevent conflicts the :func:`~nngt.Network.to_nest` function is not
        available. If you know what you are doing, you should be able to find a
        workaround...

        Parameters
        ----------
        gids : array-like
            Ids of the neurons in NEST or simply user specified ids.
        get_params : bool, optional (default: True)
            Whether the parameters should be obtained from NEST (can be very
            slow).
        neuron_model : string, optional (default: None)
            Name of the NEST neural model to use when simulating the activity.
        neuron_param : dict, optional (default: {})
            Dictionary containing the neural parameters; the default value will
            make NEST use the default parameters of the model.
        syn_model : string, optional (default: 'static_synapse')
            NEST synaptic model to use when simulating the activity.
        syn_param : dict, optional (default: {})
            Dictionary containing the synaptic parameters; the default value
            will make NEST use the default parameters of the model.

        Returns
        -------
        net : :class:`~nngt.Network` or subclass
            Uniform network of disconnected neurons.
        '''
        from nngt.lib.errors import not_implemented

        if neuron_param is None:
            neuron_param = {}
        if syn_param is None:
            syn_param = {}

        # create the population
        size  = len(gids)
        nodes = [i for i in range(size)]

        group = nngt.NeuralGroup(
            nodes, neuron_type=1, neuron_model=neuron_model,
            neuron_param=neuron_param)

        pop = nngt.NeuralPop.from_groups([group])

        # create the network
        net = cls(population=pop, **kwargs)
        net.nest_gids = np.array(gids)
        net._id_from_nest_gid = {gid: i for i, gid in enumerate(gids)}
        net.to_nest = not_implemented

        if get_connections:
            from nngt.simulation import get_nest_adjacency
            converter = {gid: i for i, gid in enumerate(gids)}
            mat = get_nest_adjacency(converter)
            edges = np.array(mat.nonzero()).T
            w = mat.data
            net.new_edges(edges, {'weight': w}, check_duplicates=False,
                          check_self_loops=False, check_existing=False)

        if get_params:
            raise NotImplementedError('`get_params` not implemented yet.')

        return net

    @classmethod
    def uniform(cls, size, neuron_model=default_neuron,
                        neuron_param=None, syn_model=default_synapse,
                        syn_param=None, **kwargs):
        '''
        Generate a network containing only one type of neurons.

        Parameters
        ----------
        size : int
            Number of neurons in the network.
        neuron_model : string, optional (default: 'aief_cond_alpha')
            Name of the NEST neural model to use when simulating the activity.
        neuron_param : dict, optional (default: {})
            Dictionary containing the neural parameters; the default value will
            make NEST use the default parameters of the model.
        syn_model : string, optional (default: 'static_synapse')
            NEST synaptic model to use when simulating the activity.
        syn_param : dict, optional (default: {})
            Dictionary containing the synaptic parameters; the default value
            will make NEST use the default parameters of the model.

        Returns
        -------
        net : :class:`~nngt.Network` or subclass
            Uniform network of disconnected neurons.
        '''
        if neuron_param is None:
            neuron_param = {}

        if syn_param is None:
            syn_param = {}

        pop = nngt.NeuralPop.uniform(
            size, neuron_model=neuron_model, neuron_param=neuron_param,
            syn_model=syn_model, syn_param=syn_param, parent=None)

        net = cls(population=pop, **kwargs)

        return net

    @classmethod
    def exc_and_inhib(cls, size, iratio=0.2, en_model=default_neuron,
            en_param=None, in_model=default_neuron, in_param=None,
            syn_spec=None, **kwargs):
        '''
        Generate a network containing a population of two neural groups:
        inhibitory and excitatory neurons.

        Parameters
        ----------
        size : int
            Number of neurons in the network.
        i_ratio : double, optional (default: 0.2)
            Ratio of inhibitory neurons: :math:`\\frac{N_i}{N_e+N_i}`.
        en_model : string, optional (default: 'aeif_cond_alpha')
           Nest model for the excitatory neuron.
        en_param : dict, optional (default: {})
            Dictionary of parameters for the the excitatory neuron.
        in_model : string, optional (default: 'aeif_cond_alpha')
           Nest model for the inhibitory neuron.
        in_param : dict, optional (default: {})
            Dictionary of parameters for the the inhibitory neuron.
        syn_spec : dict, optional (default: static synapse)
            Dictionary containg a directed edge between groups as key and the
            associated synaptic parameters for the post-synaptic neurons (i.e.
            those of the second group) as value. If provided, all connections
            between groups will be set according to the values contained in
            `syn_spec`. Valid keys are:

            - `('excitatory', 'excitatory')`
            - `('excitatory', 'inhibitory')`
            - `('inhibitory', 'excitatory')`
            - `('inhibitory', 'inhibitory')`

        Returns
        -------
        net : :class:`~nngt.Network` or subclass
            Network of disconnected excitatory and inhibitory neurons.

        See also
        --------
        :func:`~nngt.NeuralPop.exc_and_inhib`
        '''
        pop = nngt.NeuralPop.exc_and_inhib(
            size, iratio, en_model, en_param, in_model, in_param,
            syn_spec=syn_spec)

        net = cls(population=pop, **kwargs)

        return net

    #-------------------------------------------------------------------------#
    # Constructor, destructor and attributes

    def __init__(self, name="Network", weighted=True, directed=True,
                 copy_graph=None, population=None, inh_weight_factor=1.,
                 **kwargs):
        '''
        Initializes :class:`~nngt.Network` instance.

        .. versionchanged: 2.4
            Move `from_graph` to `copy_graph` to reflect changes in Graph.

        Parameters
        ----------
        nodes : int, optional (default: 0)
            Number of nodes in the graph.
        name : string, optional (default: "Graph")
            The name of this :class:`Graph` instance.
        weighted : bool, optional (default: True)
            Whether the graph edges have weight properties.
        directed : bool, optional (default: True)
            Whether the graph is directed or undirected.
        copy_graph : :class:`~nngt.core.GraphObject`, optional (default: None)
            An optional :class:`~nngt.core.GraphObject` to serve as base.
        population : :class:`nngt.NeuralPop`, (default: None)
            An object containing the neural groups and their properties:
            model(s) to use in NEST to simulate the neurons as well as their
            parameters.
        inh_weight_factor : float, optional (default: 1.)
            Factor to apply to inhibitory synapses, to compensate for example
            the strength difference due to timescales between excitatory and
            inhibitory synapses.

        Returns
        -------
        self : :class:`~nggt.Network`
        '''
        self.__id = self.__class__.__max_id

        self.__class__.__num_networks += 1
        self.__class__.__max_id += 1

        assert directed, "Network class cannot be undirected."

        if population is None:
            raise InvalidArgument("Network needs a NeuralPop to be created")

        nodes = population.size

        if "nodes" in kwargs.keys():
            assert kwargs["nodes"] == nodes, "Incompatible values for " +\
                "`nodes` = {} with a `population` of size {}.".format(
                    kwargs["nodes"], nodes)
            del kwargs["nodes"]

        if "delays" not in kwargs:  # set default delay to 1.
            kwargs["delays"] = 1.

        super().__init__(nodes=nodes, name=name, weighted=weighted,
                         directed=directed, copy_graph=copy_graph,
                         inh_weight_factor=inh_weight_factor, **kwargs)

        self._init_bioproperties(population)

    def __del__(self):
        super().__del__()
        self.__class__.__num_networks -= 1

    @property
    def population(self):
        '''
        :class:`~nngt.NeuralPop` that divides the neurons into groups with
        specific properties.
        '''
        return self._population

    @population.setter
    def population(self, population):
        if issubclass(population.__class__, nngt.NeuralPop):
            if self.node_nb() == population.size:
                if population.is_valid:
                    self._population = population
                else:
                    raise AttributeError("NeuralPop is not valid (not all "
                                         "neurons are associated to a group).")
            else:
                raise AttributeError("Network and NeuralPop must have same "
                                     "number of neurons.")
        else:
            raise AttributeError("Expecting NeuralPop but received "
                                 "'{}'".format(population.__class__.__name__))

    @property
    def nest_gids(self):
        return self._nest_gids

    @nest_gids.setter
    def nest_gids(self, gids):
        self._nest_gids = gids
        for group in self.population.values():
            group._nest_gids = gids[group.ids]

    def get_edge_types(self, edges=None):
        '''
        Return the type of all or a subset of the edges.
        For all edges, the types are ordered according to the edges ids, i.e.
        in the same order as :property:`~nngt.Graph.edges_array`.

        .. versionchanged:: 2.4
            Updated it to make it compatible with the default
            :class:`~nngt.Graph` function, including the `edges` argument.

        Parameters
        ----------
        edges : (E, 2) array, optional (default: all edges)
            Edges for which the type should be returned.

        Returns
        -------
        the list of types (1 for excitatory, -1 for inhibitory)
        '''
        edges = self.edges_array if edges is None else edges

        types = np.ones(len(edges))

        inhib_neurons = set()

        for g in self._population.values():
            if g.neuron_type == -1:
                inhib_neurons.update(g.ids)

        for i, e in enumerate(edges):
            if e[0] in inhib_neurons:
                types[i] = -1

        return types

    def id_from_nest_gid(self, gids):
        '''
        Return the ids of the nodes in the :class:`nngt.Network` instance from
        the corresponding NEST gids.

        Parameters
        ----------
        gids : int or tuple
            NEST gids.

        Returns
        -------
        ids : int or tuple
            Ids in the network. Same type as the requested `gids` type.
        '''
        if nonstring_container(gids):
            return np.array([self._id_from_nest_gid[gid] for gid in gids],
                            dtype=int)
        else:
            return self._id_from_nest_gid[gids]

    def to_nest(self, send_only=None, weights=True):
        '''
        Send the network to NEST.

        .. seealso::
            :func:`~nngt.simulation.make_nest_network` for parameters
        '''
        from nngt.simulation import make_nest_network
        if nngt._config['with_nest']:
            return make_nest_network(
                self, send_only=send_only, weights=weights)
        else:
            raise RuntimeError("NEST is not present.")

    #-------------------------------------------------------------------------#
    # Init tool

    def _init_bioproperties(self, population):
        ''' Set the population attribute and link each neuron to its group. '''
        self._population = None
        self._nest_gids = None
        self._id_from_nest_gid = None
        if not hasattr(self, '_iwf'):
            self._iwf = 1.
        if issubclass(population.__class__, nngt.NeuralPop):
            if population.is_valid or not self.node_nb():
                self._population = population
                nodes = population.size
                # create the delay attribute if necessary
                if "delay" not in self.edge_attributes:
                    self.set_delays()
            else:
                raise AttributeError("NeuralPop is not valid (not all neurons "
                                     "are associated to a group).")
        else:
            raise AttributeError("Expected NeuralPop but received "
                                 "{}".format(pop.__class__.__name__))

    #-------------------------------------------------------------------------#
    # Setter

    def set_types(self, edge_type, nodes=None, fraction=None):
        raise NotImplementedError("Cannot be used on :class:`~nngt.Network`.")

    def get_neuron_type(self, neuron_ids):
        '''
        Return the type of the neurons (+1 for excitatory, -1 for inhibitory).

        Parameters
        ----------
        neuron_ids : int or tuple
            NEST gids.

        Returns
        -------
        ids : int or tuple
            Ids in the network. Same type as the requested `gids` type.
        '''
        if is_integer(neuron_ids):
            group_name = self._population._neuron_group[neuron_ids]
            neuron_type = self._population[group_name].neuron_type
            return neuron_type
        else:
            groups = (self._population._neuron_group[i] for i in neuron_ids)
            types = tuple(self._population[gn].neuron_type for gn in groups)
            return types

    #-------------------------------------------------------------------------#
    # Getter

    def neuron_properties(self, idx_neuron):
        '''
        Properties of a neuron in the graph.

        Parameters
        ----------
        idx_neuron : int
            Index of a neuron in the graph.

        Returns
        -------
        dict of the neuron's properties.
        '''
        group_name = self._population._neuron_group[idx_neuron]
        return self._population[group_name].properties()


# -------------- #
# SpatialNetwork #
# -------------- #

class SpatialNetwork(Network, SpatialGraph):

    """
    Class that inherits from :class:`~nngt.Network` and
    :class:`~nngt.SpatialGraph` to provide a detailed description of a real
    neural network in space, i.e. with positions and biological properties to
    interact with NEST.
    """

    #-------------------------------------------------------------------------#
    # Class attributes

    __num_networks = 0
    __max_id = 0

    #-------------------------------------------------------------------------#
    # Constructor, destructor, and attributes

    def __init__(self, population, name="SpatialNetwork", weighted=True,
                 directed=True, shape=None, copy_graph=None, positions=None,
                 **kwargs):
        '''
        Initialize SpatialNetwork instance.

        .. versionchanged: 2.4
            Move `from_graph` to `copy_graph` to reflect changes in Graph.

        Parameters
        ----------
        name : string, optional (default: "Graph")
            The name of this :class:`Graph` instance.
        weighted : bool, optional (default: True)
            Whether the graph edges have weight properties.
        directed : bool, optional (default: True)
            Whether the graph is directed or undirected.
        shape : :class:`~nngt.geometry.Shape`, optional (default: None)
            Shape of the neurons' environment (None leads to a square of side
            1 cm)
        positions : :class:`numpy.array`, optional (default: None)
            Positions of the neurons; if not specified and `nodes` != 0, then
            neurons will be reparted at random inside the
            :class:`~nngt.geometry.Shape` object of the instance.
        population : class:`~nngt.NeuralPop`, optional (default: None)
            Population from which the network will be built.

        Returns
        -------
        self : :class:`~nngt.SpatialNetwork`
        '''
        self.__id = self.__class__.__max_id
        self.__class__.__num_networks += 1
        self.__class__.__max_id += 1

        if population is None:
            raise InvalidArgument("Network needs a NeuralPop to be created")

        nodes = population.size

        super().__init__(
            nodes=nodes, name=name, weighted=weighted, directed=directed,
            shape=shape, positions=positions, population=population,
            copy_graph=copy_graph, **kwargs)

    def __del__ (self):
        super().__del__()
        self.__class__.__num_networks -= 1

    #-------------------------------------------------------------------------#
    # Setter

    def set_types(self, syn_type, nodes=None, fraction=None):
        raise NotImplementedError("Cannot be used on "
                                  ":class:`~nngt.SpatialNetwork`.")



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Using complex NEST models '''

import numpy as np
import nngt
import nngt.generation as ng


# np.random.seed(0)


'''
Build a network with two populations:
* excitatory (80%)
* inhibitory (20%)
'''
num_neurons = 50   # number of neurons
avg_degree  = 20   # average number of neighbours
std_degree  = 3    # deviation for the Gaussian graph

# parameters
neuron_model = "ht_neuron"      # hill-tononi model
exc_syn = {'receptor_type': 1}  # 1 is 'AMPA' in this model
inh_syn = {'receptor_type': 3}  # 3 is 'GABA_A' in this model

synapses = {
    (1, 1):   exc_syn,
    (1, -1):  exc_syn,
    (-1, 1):  inh_syn,
    (-1, -1): inh_syn,
}

pop = nngt.NeuralPop.exc_and_inhib(
    num_neurons, en_model=neuron_model, in_model=neuron_model,
    syn_spec=synapses)

# create the network and send it to NEST
w_prop = {"distribution": "gaussian", "avg": 0.2, "std": .05}
net = nngt.generation.gaussian_degree(
    avg_degree, std_degree, population=pop, weights=w_prop)

'''
Send to NEST and set excitation and recorders
'''
if nngt.get_config('with_nest'):
    import nest
    import nngt.simulation as ns

    nest.ResetKernel()

    gids = net.to_nest()

    # add noise to the excitatory neurons
    excs = list(pop["excitatory"].nest_gids)
    inhs = list(pop["inhibitory"].nest_gids)
    ns.set_noise(excs, 10., 2.)
    ns.set_noise(inhs, 5., 1.)

    # record
    groups = [key for key in net.population]
    recorder, record = ns.monitor_groups(groups, net)

    '''
    Simulate and plot.
    '''
    simtime = 2000.
    nest.Simulate(simtime)

    if nngt.get_config('with_plot'):
        ns.plot_activity(
            recorder, record, network=net, show=True, histogram=False,
            limits=(0, simtime))



#-*- coding:utf-8 -*-
#
# analysis/activity_analysis.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Tools for activity analysis from data """

import logging

import numpy as np
import scipy.signal as sps
import scipy.sparse as ssp

from nngt.lib import nonstring_container, find_idx_nearest
from nngt.lib.logger import _log_message


__all__ = [
    "get_b2",
    "get_firing_rate",
    "get_spikes",
    "total_firing_rate",
]


logger = logging.getLogger(__name__)


# ----------------------- #
# Get activity properties #
# ----------------------- #

def get_b2(network=None, spike_detector=None, data=None, nodes=None):
    '''
    Return the B2 coefficient for the neurons.

    Parameters
    ----------
    network : :class:`nngt.Network`, optional (default: None)
        Network for which the activity was simulated.
    spike_detector : tuple of ints, optional (default: spike detectors)
        GID of the "spike_detector" objects recording the network activity.
    data : array-like of shape (N, 2), optionale (default: None)
        Array containing the spikes data (first line must contain the NEST GID
        of the neuron that fired, second line must contain the associated spike
        time).
    nodes : array-like, optional (default: all neurons)
        NNGT ids of the nodes for which the B2 should be computed.

    Returns
    -------
    b2 : array-like
        B2 coefficient for each neuron in `nodes`.
    '''
    if data is None:
        data, nodes = _set_data_nodes(network, data, nodes)
        data = _set_spike_data(data, spike_detector)
    else:
        if nodes is None:
            nodes = np.unique(data[:, 0])

    return _b2_from_data(nodes, data)


def get_firing_rate(network=None, spike_detector=None, data=None, nodes=None):
    '''
    Return the average firing rate for the neurons.

    Parameters
    ----------
    network : :class:`nngt.Network`, optional (default: None)
        Network for which the activity was simulated.
    spike_detector : tuple of ints, optional (default: spike detectors)
        GID of the "spike_detector" objects recording the network activity.
    data : :class:`numpy.array` of shape (N, 2), optionale (default: None)
        Array containing the spikes data (first line must contain the NEST GID
        of the neuron that fired, second line must contain the associated spike
        time).
    nodes : array-like, optional (default: all nodes)
        NNGT ids of the nodes for which the B2 should be computed.

    Returns
    -------
    fr : array-like
        Firing rate for each neuron in `nodes`.
    '''
    if data is None:
        data, nodes = _set_data_nodes(network, data, nodes)
        data = _set_spike_data(data, spike_detector)
    else:
        if nodes is None:
            nodes = np.unique(data[:, 0])

    return _fr_from_data(nodes, data)


def total_firing_rate(network=None, spike_detector=None, nodes=None, data=None,
                      kernel_center=0., kernel_std=30., resolution=None,
                      cut_gaussian=5.):
    '''
    Computes the total firing rate of the network from the spike times.
    Firing rate is obtained as the convolution of the spikes with a Gaussian
    kernel characterized by a standard deviation and a temporal shift.

    .. versionadded:: 0.7

    Parameters
    ----------
    network : :class:`nngt.Network`, optional (default: None)
        Network for which the activity was simulated.
    spike_detector : tuple of ints, optional (default: spike detectors)
        GID of the "spike_detector" objects recording the network activity.
    data : :class:`numpy.array` of shape (N, 2), optionale (default: None)
        Array containing the spikes data (first line must contain the NEST GID
        of the neuron that fired, second line must contain the associated spike
        time).
    kernel_center : float, optional (default: 0.)
        Temporal shift of the Gaussian kernel, in ms.
    kernel_std : float, optional (default: 30.)
        Characteristic width of the Gaussian kernel (standard deviation) in ms.
    resolution : float or array, optional (default: `0.1*kernel_std`)
        The resolution at which the firing rate values will be computed.
        Choosing a value smaller than `kernel_std` is strongly advised.
        If resolution is an array, it will be considered as the times were the
        firing rate should be computed.
    cut_gaussian : float, optional (default: 5.)
        Range over which the Gaussian will be computed. By default, we consider
        the 5-sigma range. Decreasing this value will increase speed at the
        cost of lower fidelity; increasing it with increase the fidelity at the
        cost of speed.

    Returns
    -------
    fr : array-like
        The firing rate in Hz.
    times : array-like
        The times associated to the firing rate values.
    '''
    times, kernel_size = None, None

    if data is None:
        data, _ = _set_data_nodes(network, data, nodes)
        data    = _set_spike_data(data, spike_detector)

    # set resolution and kernel properties + generate the times
    if resolution is None:
        resolution = 0.1*kernel_std

    if nonstring_container(resolution):
        dt = np.diff(resolution)
        assert np.allclose(dt - dt[0], 0.), 'If `resolution` is an array, ' +\
                                            'it must contain evenly spaced ' +\
                                            'times.'
        times = np.array(resolution)
        resolution = dt[0]

    bin_std = int(kernel_std / float(resolution))
    kernel_size = int(2. * cut_gaussian * bin_std)

    if times is None:
        delta_T = resolution * 0.5 * kernel_size
        times = np.arange(np.min(data[:, 1]) - delta_T,
                          np.max(data[:, 1]) + delta_T, resolution)

    rate = np.zeros(len(times))

    # counts the spikes at each time
    pos = find_idx_nearest(times, data[:, 1])
    bins = np.linspace(0, len(times), len(times)+1)
    counts, _ = np.histogram(pos, bins=bins)

    # initialize with delta rate in Hz
    rate += 1000. * counts / (kernel_std*np.sqrt(np.pi))
    fr = _smooth(rate, kernel_size, bin_std, mode='same')

    # translate times
    times += kernel_center

    return fr, times


def get_spikes(recorder=None, spike_times=None, senders=None, astype="ssp"):
    '''
    Return a 2D sparse matrix, where:

    - each row i contains the spikes of neuron i (in NEST),
    - each column j contains the times of the jth spike for all neurons.

    .. versionchanged:: 1.0
        Neurons are now located in the row corresponding to their NEST GID.

    Parameters
    ----------
    recorder : tuple, optional (default: None)
        Tuple of NEST gids, where the first one should point to the
        spike_detector which recorded the spikes.
    spike_times : array-like, optional (default: None)
        If `recorder` is not provided, the spikes' data can be passed directly
        through their `spike_times` and the associated `senders`.
    senders : array-like, optional (default: None)
        `senders[i]` corresponds to the neuron which fired at `spike_times[i]`.
    astype : str, optional (default: "ssp")
        Format of the returned data. Default is sparse lil_matrix ("ssp")
        with one row per neuron, otherwise "np" returns a (T, 2) array, with
        T the number of spikes (the first row being the NEST gid, the second
        the spike time).

    Example
    -------
    >>> get_spikes()

    >>> get_spikes(recorder)

    >>> times = [1.5, 2.68, 125.6]
    >>> neuron_ids = [12, 0, 65]
    >>> get_spikes(spike_times=times, senders=neuron_ids)

    Note
    ----
    If no arguments are passed to the function, the first spike_recorder
    available in NEST will be used.
    Neuron positions correspond to their GIDs in NEST.

    Returns
    -------
    CSR matrix containing the spikes sorted by neuron GIDs (rows) and time
    (columns).
    '''
    if recorder is not None:
        from ..simulation.nest_utils import _get_nest_gids
        import nest
        data = nest.GetStatus(_get_nest_gids(recorder))[0]["events"]
        spike_times = data["times"]
        senders = data["senders"]
    elif spike_times is None and senders is None:
        from ..simulation.nest_utils import _get_nest_gids, spike_rec
        import nest
        nodes = nest.GetNodes(properties={'model': spike_rec})
        data = nest.GetStatus(nodes)[0]["events"]
        spike_times = data["times"]
        senders = data["senders"]

    if astype == "np":
        return np.array([senders, spike_times]).T
    elif astype == "ssp":
        if np.any(senders):
            max_sender = np.max(senders)
            # create the sparse matrix
            data = [0 for _ in range(max_sender + 1)]
            row_idx = []
            col_idx = []
            for time, neuron in zip(spike_times, senders):
                row_idx.append(neuron)
                col_idx.append(data[neuron])
                data[neuron] += 1
            return ssp.csr_matrix((spike_times, (row_idx, col_idx)))
        else:
            return ssp.csr_matrix([])


# ----- #
# Tools #
# ----- #

def _b2_from_data(ids, data):
    b2 = np.full(len(ids), np.NaN)
    if len(data[:, 0]) > 0:
        for i, neuron in enumerate(ids):
            ids = np.where(data[:, 0] == neuron)[0]
            dt1 = np.diff(data[ids, 1])
            dt2 = dt1[1:] + dt1[:-1]
            avg_isi = np.mean(dt1)
            if avg_isi != 0.:
                b2[i] = (2*np.var(dt1) - np.var(dt2)) / (2*avg_isi**2)
            else:
                b2[i] = np.inf
    else:
        _log_message(logger, "WARNING", 'No spikes in the data.')
    return b2


def _fr_from_data(ids, data):
    fr = np.zeros(len(ids))

    if len(data[:, 0]):
        T = float(np.max(data[:, 1]) - np.min(data[:, 1]))

        for i, neuron in enumerate(ids):
            ids = np.where(data[:, 0] == neuron)[0]
            fr[i] = len(ids) / T

    return fr


def _set_data_nodes(network, data, nodes):
    from ..simulation.nest_utils import _get_nest_gids

    if data is None:
        data = [[], []]

    if nodes is None:
        nodes = network.nest_gids
    else:
        nodes = network.nest_gids[nodes]

    return data, _get_nest_gids(nodes)


def _set_spike_data(data, spike_detector):
    '''
    Data must be [[], []]
    '''
    import nest
    from ..simulation.nest_utils import _get_nest_gids, spike_rec, nest_version

    if not len(data[0]):
        if spike_detector is None:
            prop = {'model': spike_rec}
            if nest_version == 3:
                spike_detector = nest.GetNodes(properties=prop)
            else:
                spike_detector = nest.GetNodes((0,), properties=prop)[0]

        events = nest.GetStatus(spike_detector, "events")

        for ev_dict in events:
            data[0].extend(ev_dict["senders"])
            data[1].extend(ev_dict["times"])

    sorter = np.argsort(data[1])

    return np.array(data)[:, sorter].T


def _smooth(data, kernel_size, std, mode='same'):
    '''
    Convolve an array by a Gaussian kernel.

    Parameters
    ----------
    kernel_size : int
        Size of the kernel array in bins.
    std : float
        Width of the Gaussian (also in bins).

    Returns
    -------
    convolved array.
    '''
    kernel = sps.gaussian(kernel_size, std)
    kernel /= np.sum(kernel)
    return sps.convolve(data, kernel, mode=mode)



#-*- coding:utf-8 -*-
#
# core/neural_pop_group.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Graph data strctures in NNGT """

import logging
import weakref
from copy import deepcopy

import numpy as np

import nngt
from nngt.lib import (InvalidArgument, nonstring_container, is_integer,
                      default_neuron, default_synapse)
from nngt.lib._frozendict import _frozendict
from nngt.lib.logger import _log_message

from .group_structure import Structure, Group, MetaGroup


__all__ = [
    'GroupProperty',
    'MetaNeuralGroup',
    'NeuralGroup',
    'NeuralPop',
]

logger = logging.getLogger(__name__)


# --------- #
# NeuralPop #
# --------- #

class NeuralPop(Structure):

    """
    The basic class that contains groups of neurons and their properties.

    :ivar has_models: :obj:`bool`,
        ``True`` if every group has a ``model`` attribute.
    :ivar ~nngt.NeuralPop.size: :obj:`int`,
        Returns the number of neurons in the population.
    :ivar syn_spec: :obj:`dict`,
        Dictionary containing informations about the synapses between the
        different groups in the population.
    :ivar ~nngt.NeuralPop.is_valid: :obj:`bool`,
        Whether this population can be used to create a network in NEST.
    """

    # number of created populations
    __num_created = 0

    # store weakrefs to created populations
    __pops = weakref.WeakValueDictionary()

    #-------------------------------------------------------------------------#
    # Class attributes and methods

    @classmethod
    def from_network(cls, graph, *args):
        '''
        Make a NeuralPop object from a network. The groups of neurons are
        determined using instructions from an arbitrary number of
        :class:`~nngt.properties.GroupProperties`.
        '''
        return cls(parent=graph, graph=graph, group_prop=args)

    @classmethod
    def from_groups(cls, groups, names=None, syn_spec=None, parent=None,
                    meta_groups=None, with_models=True):
        '''
        Make a NeuralPop object from a (list of) :class:`~nngt.NeuralGroup`
        object(s).

        Parameters
        ----------
        groups : list of :class:`~nngt.NeuralGroup` objects
            Groups that will be used to form the population. Note that a given
            neuron can only belong to a single group, so the groups should form
            pairwise disjoints complementary sets.
        names : list of str, optional (default: None)
            Names that can be used as keys to retreive a specific group. If not
            provided, keys will be the group name (if not empty) or the position
            of the group in `groups`, stored as a string.
            In the latter case, the first group in a population named `pop`
            will be retreived by either `pop[0]` or `pop['0']`.
        parent : :class:`~nngt.Graph`, optional (default: None)
            Parent if the population is created from an exiting graph.
        syn_spec : dict, optional (default: static synapse)
            Dictionary containg a directed edge between groups as key and the
            associated synaptic parameters for the post-synaptic neurons (i.e.
            those of the second group) as value.
            If a 'default' entry is provided, all unspecified connections will
            be set to its value.
        meta_groups : list or dict of str/:class:`~nngt.NeuralGroup` items
            Additional set of groups which can overlap: a neuron can belong to
            several different meta groups. Contrary to the primary groups, meta
            groups do therefore no need to be disjoint.
            If all meta-groups have a name, they can be passed directly through
            a list; otherwise a dict is necessary.
        with_model : bool, optional (default: True)
            Whether the groups require models (set to False to use populations
            for graph theoretical purposes, without NEST interaction)

        Example
        -------
        For synaptic properties, if provided in `syn_spec`, all connections
        between groups will be set according to the values.
        Keys can be either group names or types (1 for excitatory, -1 for
        inhibitory). Because of this, several combination can be available for
        the connections between two groups. Because of this, priority is given
        to source (presynaptic properties), i.e. NNGT will look for the entry
        matching the first group name as source before looking for entries
        matching the second group name as target.

        .. code-block:: python

            # we created groups `g1`, `g2`, and `g3`
            prop = {
                ('g1', 'g2'): {'model': 'tsodyks2_synapse', 'tau_fac': 50.},
                ('g1', g3'): {'weight': 100.},
                ...
            }
            pop = NeuronalPop.from_groups(
                [g1, g2, g3], names=['g1', 'g2', 'g3'], syn_spec=prop)

        Note
        ----
        If the population is not generated from an existing
        :class:`~nngt.Graph` and the groups do not contain explicit ids, then
        the ids will be generated upon population creation: the first group, of
        size N0, will be associated the indices 0 to N0 - 1, the second group
        (size N1), will get N0 to N0 + N1 - 1, etc.
        '''
        if not nonstring_container(groups):
            groups = [groups]

        gsize = len(groups)
        names = [] if names is None else list(names)

        if not names:
            for i, g in enumerate(groups):
                if g.name:
                    names.append(g.name)
                else:
                    names.append(str(i))

        assert len(names) == gsize, "`names` and `groups` must have " +\
                                    "the same size."

        for n in names:
            assert isinstance(n, str), "Group names must be strings."

        if syn_spec:
            _check_syn_spec(syn_spec, names, groups)

        current_size = 0
        for g in groups:
            # generate the neuron ids if necessary
            ids = g.ids
            if len(ids) == 0:
                ids = list(range(current_size, current_size + g.size))
                g.ids = ids
            current_size += len(ids)

        pop = cls(current_size, parent=parent, meta_groups=meta_groups,
                  with_models=with_models)

        for name, g in zip(names, groups):
            pop[name] = g
            g._struct = weakref.ref(pop)
            g._net    = weakref.ref(parent) if parent is not None else None

        # take care of synaptic connections
        pop._syn_spec = deepcopy(syn_spec if syn_spec is not None else {})

        return pop

    @classmethod
    def uniform(cls, size, neuron_type=1, neuron_model=default_neuron,
                neuron_param=None, syn_model=default_synapse, syn_param=None,
                parent=None, meta_groups=None):
        '''
        Make a NeuralPop of identical neurons belonging to a single "default"
        group.

        .. versionchanged:: 1.2
            Added `neuron_type` and `meta_groups` parameters

        Parameters
        ----------
        size : int
            Number of neurons in the population.
        neuron_type : int, optional (default: 1)
            Type of the neurons in the population: 1 for excitatory or -1 for
            inhibitory.
        neuron_model : str, optional (default: default neuron model)
            Neuronal model for the simulator.
        neuron_param : dict, optional (default: default neuron parameters)
            Parameters associated to `neuron_model`.
        syn_model : str, optional (default: default static synapse)
            Synapse model for the simulator.
        syn_param : dict, optional (default: default synaptic parameters)
            Parameters associated to `syn_model`.
        parent : :class:`~nngt.Graph` object, optional (default: None)
            Parent graph described by the population.
        meta_groups : list or dict of str/:class:`~nngt.NeuralGroup` items
            Set of groups which can overlap: a neuron can belong to
            several different meta groups, i.e. they do no need to be disjoint.
            If all meta-groups have a name, they can be passed directly through
            a list; otherwise a dict is necessary.
        '''
        neuron_param = {} if neuron_param is None else neuron_param.copy()

        if syn_param is not None:
            assert 'weight' not in syn_param, '`weight` cannot be set here.'
            assert 'delay' not in syn_param, '`delay` cannot be set here.'
            syn_param = syn_param.copy()
        else:
            syn_param = {}

        pop = cls(size, parent, meta_groups=meta_groups)
        pop.create_group(range(size), "default", neuron_type, neuron_model,
                         neuron_param)

        pop._syn_spec = {'model': syn_model}

        if syn_param is not None:
            pop._syn_spec.update(syn_param)

        return pop

    @classmethod
    def exc_and_inhib(cls, size, iratio=0.2, en_model=default_neuron,
                      en_param=None, in_model=default_neuron, in_param=None,
                      syn_spec=None, parent=None, meta_groups=None):
        '''
        Make a NeuralPop with a given ratio of inhibitory and excitatory
        neurons.

        .. versionchanged:: 0.8
            Added `syn_spec` parameter.

        .. versionchanged:: 1.2
            Added `meta_groups` parameter

        Parameters
        ----------
        size : int
            Number of neurons contained by the population.
        iratio : float, optional (default: 0.2)
            Fraction of the neurons that will be inhibitory.
        en_model : str, optional (default: default_neuron)
            Name of the NEST model that will be used to describe excitatory
            neurons.
        en_param : dict, optional (default: default NEST parameters)
            Parameters of the excitatory neuron model.
        in_model : str, optional (default: default_neuron)
            Name of the NEST model that will be used to describe inhibitory
            neurons.
        in_param : dict, optional (default: default NEST parameters)
            Parameters of the inhibitory neuron model.
        syn_spec : dict, optional (default: static synapse)
            Dictionary containg a directed edge between groups as key and the
            associated synaptic parameters for the post-synaptic neurons (i.e.
            those of the second group) as value. If provided, all connections
            between groups will be set according to the values contained in
            `syn_spec`. Valid keys are:

            - `('excitatory', 'excitatory')`
            - `('excitatory', 'inhibitory')`
            - `('inhibitory', 'excitatory')`
            - `('inhibitory', 'inhibitory')`
        parent : :class:`~nngt.Network`, optional (default: None)
            Network associated to this population.
        meta_groups : list dict of str/:class:`~nngt.NeuralGroup` items
            Additional set of groups which can overlap: a neuron can belong to
            several different meta groups. Contrary to the primary 'excitatory'
            and 'inhibitory' groups, meta groups are therefore no necessarily
            disjoint.
            If all meta-groups have a name, they can be passed directly through
            a list; otherwise a dict is necessary.

        See also
        --------
        :func:`nest.Connect` for a description of the dict that can be passed
        as values for the `syn_spec` parameter.
        '''
        num_exc_neurons = int(size*(1-iratio))

        pop = cls(size, parent, meta_groups=meta_groups)

        pop.create_group(
            range(num_exc_neurons), "excitatory", neuron_type=1,
            neuron_model=en_model, neuron_param=en_param)
        pop.create_group(
            range(num_exc_neurons, size), "inhibitory", neuron_type=-1,
            neuron_model=in_model, neuron_param=in_param)

        if syn_spec:
            _check_syn_spec(
                syn_spec, ["excitatory", "inhibitory"], pop.values())
            pop._syn_spec = deepcopy(syn_spec)
        else:
            pop._syn_spec = {}

        return pop

    @classmethod
    def _nest_reset(cls):
        '''
        Reset the _to_nest bool and potential parent networks.
        '''
        for pop in cls.__pops.valuerefs():
            if pop() is not None:
                pop()._to_nest = False
                for g in pop().values():
                    g._to_nest = False
                if pop().parent is not None:
                    pop().parent._nest_gids = None

    #-------------------------------------------------------------------------#
    # Contructor and instance attributes

    def __init__(self, size=None, parent=None, meta_groups=None,
                 with_models=True, **kwargs):
        '''
        Initialize NeuralPop instance.

        Parameters
        ----------
        size : int, optional (default: 0)
            Number of neurons that the population will contain.
        parent : :class:`~nngt.Network`, optional (default: None)
            Network associated to this population.
        meta_groups : dict of str/:class:`~nngt.NeuralGroup` items
            Optional set of groups. Contrary to the primary groups which
            define the population and must be disjoint, meta groups can
            overlap: a neuron can belong to several different meta
            groups.
        with_models : :class:`bool`
            whether the population's groups contain models to use in NEST
        *args : items for OrderedDict parent
        **kwargs : :obj:`dict`

        Returns
        -------
        pop : :class:`~nngt.NeuralPop` object.
        '''
        super().__init__(size=size, parent=parent, meta_groups=meta_groups,
                         **kwargs)

        self._syn_spec = {}
        self._has_models = with_models

        # whether the network this population represents was sent to NEST
        self._to_nest = False

        # update class properties
        self.__id = self.__class__.__num_created
        self.__class__.__num_created += 1
        self.__class__.__pops[self.__id] = self

    def __reduce__(self):
        '''
        Overwrite this function to make NeuralPop pickable.
        OrderedDict.__reduce__ returns a 3 to 5 tuple:
        - the first is the class
        - the second is the init args in Py2, empty sequence in Py3
        - the third can be used to store attributes
        - the fourth is None and needs to stay None
        - the last must be kept unchanged: odict_iterator in Py3
        '''
        state    = super().__reduce__()
        newstate = (
            NeuralPop, state[1][:3] + (self._has_models,) + state[1][3:],
            state[2], state[3], state[4]
        )

        return newstate

    def __setitem__(self, key, value):
        if self._to_nest:
            raise RuntimeError("Populations items can no longer be modified "
                               "once the network has been sent to NEST!")
        super().__setitem__(key, value)

    def copy(self):
        '''
        Return a deep copy of the population.
        '''
        # copy groups and metagroups
        groups = {k: v.copy() for k, v in self.items()}
        metagroups = {k: v.copy() for k, v in self._meta_groups.items()}

        # generate new population
        copy = NeuralPop.from_groups(
            groups.values(), groups.keys(), syn_spec=self._syn_spec,
            parent=None, meta_groups=metagroups, with_models=self._has_models)

        return copy

    @property
    def nest_gids(self):
        '''
        Return the NEST gids of the nodes inside the population.

        .. versionadded:: 1.3
        '''
        gids = []

        for g in self.values():
            gids.extend(g.nest_gids)

        return gids

    @property
    def excitatory(self):
        '''
        Return the ids of all excitatory nodes inside the population.

        .. versionadded:: 1.3
        '''
        ids = []

        for g in self.values():
            if g.neuron_type == 1:
                ids.extend(g.ids)

        return ids

    @property
    def inhibitory(self):
        '''
        Return the ids of all inhibitory nodes inside the population.

        .. versionadded:: 1.3
        '''
        ids = []

        for g in self.values():
            if g.neuron_type == -1:
                ids.extend(g.ids)

        return ids

    @property
    def syn_spec(self):
        '''
        The properties of the synaptic connections between groups.
        Returns a :obj:`dict` containing tuples as keys and dicts of parameters
        as values.

        The keys are tuples containing the names of the groups in the
        population, with the projecting group first (presynaptic neurons) and
        the receiving group last (post-synaptic neurons).

        Example
        -------
        For a population of excitatory ("exc") and inhibitory ("inh") neurons.

        .. code-block:: python

            syn_spec = {
                ("exc", "exc"): {'model': 'stdp_synapse', 'weight': 2.5},
                ("exc", "inh"): {'model': 'static_synapse'},
                ("exc", "inh"): {'model': 'stdp_synapse', 'delay': 5.},
                ("inh", "inh"): {
                    'model': 'stdp_synapse', 'weight': 5.,
                    'delay': ('normal', 5., 2.)}
                }
            }

        .. versionadded:: 0.8
        '''
        return deepcopy(self._syn_spec)

    @syn_spec.setter
    def syn_spec(self, syn_spec):
        raise NotImplementedError('`syn_spec` is not settable yet.')

    @property
    def has_models(self):
        return self._has_models

    #-------------------------------------------------------------------------#
    # Methods

    def create_group(self, neurons, name, neuron_type=1, neuron_model=None,
                     neuron_param=None, replace=False):
        '''
        Create a new group in the population.

        Parameters
        ----------
        neurons : int or array-like
            Desired number of neurons or list of the neurons indices.
        name : str
            Name of the group.
        neuron_type : int, optional (default: 1)
            Type of the neurons : 1 for excitatory, -1 for inhibitory.
        neuron_model : str, optional (default: None)
            Name of a neuron model in NEST.
        neuron_param : dict, optional (default: None)
            Parameters for `neuron_model` in the NEST simulator. If None,
            default parameters will be used.
        replace : bool, optional (default: False)
            Whether to override previous exiting meta group with same name.
        '''
        assert isinstance(name, str), "Group `name` must be a string."
        assert neuron_type in (-1, 1), "Valid neuron type must be -1 or 1."

        if self._to_nest:
            raise RuntimeError("Groups can no longer be created once the "
                               "network has been sent to NEST!")

        if name in self and not replace:
            raise KeyError("Group with name '" + name + "' already " +\
                           "exists. Use `replace=True` to overwrite it.")

        neuron_param = {} if neuron_param is None else neuron_param.copy()

        group = NeuralGroup(neurons, neuron_type=neuron_type,
                            neuron_model=neuron_model,
                            neuron_param=neuron_param, name=name)

        self[name] = group

    def create_meta_group(self, neurons, name, neuron_param=None,
                          replace=False):
        '''
        Create a new meta group and add it to the population.

        Parameters
        ----------
        neurons : int or array-like
            Desired number of neurons or list of the neurons indices.
        name : str
            Name of the group.
        neuron_type : int, optional (default: 1)
            Type of the neurons : 1 for excitatory, -1 for inhibitory.
        neuron_model : str, optional (default: None)
            Name of a neuron model in NEST.
        neuron_param : dict, optional (default: None)
            Parameters for `neuron_model` in the NEST simulator. If None,
            default parameters will be used.
        replace : bool, optional (default: False)
            Whether to override previous exiting meta group with same name.
        '''
        neuron_param = {} if neuron_param is None else neuron_param.copy()

        group = MetaNeuralGroup(neurons, name=name, neuron_param=neuron_param)

        self.add_meta_group(group, replace=replace)

        return group

    def set_model(self, model, group=None):
        '''
        Set the groups' models.

        Parameters
        ----------
        model : dict
            Dictionary containing the model type as key ("neuron" or "synapse")
            and the model name as value (e.g. {"neuron": "iaf_neuron"}).
        group : list of strings, optional (default: None)
            List of strings containing the names of the groups which models
            should be updated.

        Note
        ----
        By default, synapses are registered as "static_synapse"s in NEST;
        because of this, only the ``neuron_model`` attribute is checked by
        the ``has_models`` function: it will answer ``True`` if all groups
        have a 'non-None' ``neuron_model`` attribute.

        Warning
        -------
        No check is performed on the validity of the models, which means
        that errors will only be detected when building the graph in NEST.
        '''
        if self._to_nest:
            raise RuntimeError("Models cannot be changed after the network "
                               "has been sent to NEST!")
        if group is None:
            group = self.keys()
        try:
            for key, val in model.items():
                for name in group:
                    if key == "neuron":
                        self[name].neuron_model = val
                    elif key == "synapse":
                        self[name].syn_model = val
                    else:
                        raise ValueError(
                            "Model type {} is not valid; choose among 'neuron'"
                            " or 'synapse'.".format(key))
        except:
            if model is not None:
                raise InvalidArgument(
                    "Invalid model dict or group; see docstring.")

        b_has_models = True

        if model is None:
            b_has_models = False

        for group in self.values():
            b_has_models *= group.has_model

        self._has_models = b_has_models

    def set_neuron_param(self, params, neurons=None, group=None):
        '''
        Set the parameters of specific neurons or of a whole group.

        .. versionadded:: 1.0

        Parameters
        ----------
        params : dict
            Dictionary containing parameters for the neurons. Entries can be
            either a single number (same for all neurons) or a list (one entry
            per neuron).
        neurons : list of ints, optional (default: None)
            Ids of the neurons whose parameters should be modified.
        group : list of strings, optional (default: None)
            List of strings containing the names of the groups whose parameters
            should be updated. When modifying neurons from a single group, it
            is still usefull to specify the group name to speed up the pace.

        Note
        ----
        If both `neurons` and `group` are None, all neurons will be modified.

        Warning
        -------
        No check is performed on the validity of the parameters, which means
        that errors will only be detected when building the graph in NEST.
        '''
        if self._to_nest:
            raise RuntimeError("Parameters cannot be changed after the "
                               "network has been sent to NEST!")

        if neurons is not None:  # specific neuron ids
            groups = []
            # get the groups they could belong to
            if group is not None:
                if nonstring_container(group):
                    groups.extend((self[g] for g in group))
                else:
                    groups.append(self[group])
            else:
                groups.extend(self.values())
            # update the groups parameters
            for g in groups:
                idx = np.where(np.in1d(g.ids, neurons, assume_unique=True))[0]
                # set the properties of the nodes for each entry in params
                for k, v in params.items():
                    default = np.NaN
                    if k in g.neuron_param:
                        default = g.neuron_param[k]
                    elif nngt.get_config('with_nest'):
                        try:
                            import nest
                            try:
                                default = nest.GetDefaults(g.neuron_model, k)
                            except nest.NESTError:
                                pass
                        except ImportError:
                            pass
                    vv      = np.repeat(default, g.size)
                    vv[idx] = v
                    # update
                    g.neuron_param[k] = vv
        else:  # all neurons in one or several groups
            group = self.keys() if group is None else group
            if not nonstring_container(group):
                group = [group]
            start = 0
            for name in group:
                g = self[name]
                for k, v in params.items():
                    if nonstring_container(v):
                        g.neuron_param[k] = v[start:start+g.size]
                    else:
                        g.neuron_param[k] = v
                start += g.size

    def get_param(self, groups=None, neurons=None, element="neuron"):
        '''
        Return the `element` (neuron or synapse) parameters for neurons or
        groups of neurons in the population.

        Parameters
        ----------
        groups : ``str``, ``int`` or array-like, optional (default: ``None``)
            Names or numbers of the groups for which the neural properties
            should be returned.
        neurons : int or array-like, optional (default: ``None``)
            IDs of the neurons for which parameters should be returned.
        element : ``list`` of ``str``, optional (default: ``"neuron"``)
            Element for which the parameters should be returned (either
            ``"neuron"`` or ``"synapse"``).

        Returns
        -------
        param : ``list``
            List of all dictionaries with the elements' parameters.
        '''
        if neurons is not None:
            groups = self._neuron_group[neurons]
        elif groups is None:
            groups = tuple(self.keys())
        key = "neuron_param" if element == "neuron" else "syn_param"
        if isinstance(groups, (str, int, np.integer)):
            return self[groups].properties[key]
        else:
            param = []
            for group in groups:
                param.append(self[group].properties[key])
            return param

    def add_to_group(self, group_name, ids):
        '''
        Add neurons to a specific group.

        Parameters
        ----------
        group_name : str or int
            Name or index of the group.
        ids : list or 1D-array
            Neuron ids.
        '''
        if self._to_nest:
            raise RuntimeError("Groups cannot be changed after the "
                               "network has been sent to NEST!")
        super().add_to_group(group_name, ids)

    def _validity_check(self, name, group):
        if self._has_models and not group.has_model:
            raise AttributeError(
                "This NeuralPop requires group to have a model attribute that "
                "is not `None`; to disable this, use `set_model(None)` "
                "method on this NeuralPop instance or set `with_models` to "
                "False when creating it.")
        elif group.has_model and not self._has_models:
            _log_message(logger, "WARNING",
                         "This NeuralPop is not set to take models into "
                         "account; use the `set_model` method to change its "
                         "behaviour.")

        if group.neuron_type not in (-1, 1):
            raise AttributeError("Valid neuron type must be -1 or 1.")

        # check pairwise disjoint
        super()._validity_check(name, group)

    def _sent_to_nest(self):
        '''
        Signify to the population and its groups that the network was sent
        to NEST and that therefore properties and groups should no longer
        be modified.
        '''
        self._to_nest = True

        for g in self.values():
            g._to_nest = True


# ----------------------------- #
# NeuralGroup and GroupProperty #
# ----------------------------- #

class NeuralGroup(Group):

    """
    Class defining groups of neurons.

    Its main variables are:

    :ivar ~nngt.NeuralGroup.ids: :obj:`list` of :obj:`int`
        the ids of the neurons in this group.
    :ivar ~nngt.NeuralGroup.neuron_type: :obj:`int`
        the default is ``1`` for excitatory neurons; ``-1`` is for inhibitory
        neurons; meta-groups must have `neuron_type` set to ``None``
    :ivar ~nngt.NeuralGroup.neuron_model: str, optional (default: None)
        the name of the model to use when simulating the activity of this group
    :ivar ~nngt.NeuralGroup.neuron_param: dict, optional (default: {})
        the parameters to use (if they differ from the model's defaults)
    :ivar ~nngt.NeuralGroup.is_metagroup: :obj:`bool`
        whether the group is a meta-group or not (`neuron_type` is ``None``
        for meta-groups)

    Warning
    -------
    Equality between :class:`~nngt.properties.NeuralGroup`s only compares
    the  size and neuronal type, ``model`` and ``param`` attributes.
    This means that groups differing only by their ``ids`` will register as
    equal.
    """

    __num_created = 0

    def __new__(cls, *args, **kwargs):
        # check neuron type for MetaGroup
        neuron_type = 1 if cls == NeuralGroup else None

        if "neuron_type" in kwargs:
            neuron_type = kwargs["neuron_type"]
        elif len(args) > 1 and is_integer(args[1]):
            neuron_type = arg[1]

        metagroup = (neuron_type is None)

        kwargs["metagroup"] = metagroup

        obj = super().__new__(cls, *args, **kwargs)

        if metagroup:
            obj.__class__ = nngt.MetaNeuralGroup

        return obj

    def __init__(self, nodes=None, neuron_type=1, neuron_model=None,
                 neuron_param=None, name=None, **kwargs):
        '''
        Calling the class creates a group of neurons.
        The default is an empty group but it is not a valid object for
        most use cases.

        Parameters
        ----------
        nodes : int or array-like, optional (default: None)
            Desired size of the group or, a posteriori, NNGT indices of the
            neurons in an existing graph.
        neuron_type : int, optional (default: 1)
            Type of the neurons (1 for excitatory, -1 for inhibitory) or None
            if not relevant (only allowed for metag roups).
        neuron_model : str, optional (default: None)
            NEST model for the neuron.
        neuron_param : dict, optional (default: model defaults)
            Dictionary containing the parameters associated to the NEST model.

        Returns
        -------
        A new :class:`~nngt.core.NeuralGroup` instance.
        '''
        super().__init__(nodes, **kwargs)

        assert neuron_type in (1, -1, None), \
            "`neuron_type` can either be 1 or -1."

        neuron_param = {} if neuron_param is None else neuron_param.copy()

        self._has_model = False if neuron_model is None else True
        self._neuron_model = neuron_model

        group_num  = NeuralGroup.__num_created + 1
        self._name = "Group {}".format(group_num) if name is None \
                                                  else name

        self._nest_gids = None
        self._neuron_param = neuron_param if self._has_model else {}
        self._neuron_type = neuron_type

        # whether the network this group belongs to was sent to NEST
        self._to_nest = False

        # parents
        self._struct = None
        self._net    = None

        NeuralGroup.__num_created += 1

    def __eq__ (self, other):
        if isinstance(other, NeuralGroup):
            same_size = self.size == other.size
            same_nmodel = ((self.neuron_model == other.neuron_model)
                           * (self.neuron_param == other.neuron_param))
            same_type = self.neuron_type == other.neuron_type

            return same_size*same_nmodel*same_type

        return False

    def __str__(self):
        return "NeuralGroup({}size={})".format(
            self._name + ": " if self._name else "", self.size)

    def _repr_pretty_(self, p, cycle):
        return p.text(str(self))

    def copy(self):
        '''
        Return a deep copy of the group.
        '''
        copy = NeuralGroup(nodes=self._ids, neuron_type=self._neuron_type,
                           neuron_model=self._neuron_model,
                           neuron_param=self._neuron_param, name=self._name)

        return copy

    @property
    def neuron_model(self):
        return self._neuron_model

    @property
    def neuron_type(self):
        return self._neuron_type

    @neuron_model.setter
    def neuron_model(self, value):
        if self._to_nest:
            raise RuntimeError("Models cannot be changed after the "
                               "network has been sent to NEST!")
        self._neuron_model = value
        self._has_model = False if value is None else True

    @property
    def neuron_param(self):
        if self._to_nest:
            return _frozendict(self._neuron_param, message="Cannot set " +
                               "neuron params after the network has been " +
                               "sent to NEST!")
        else:
            return self._neuron_param

    @neuron_param.setter
    def neuron_param(self, value):
        if self._to_nest:
            raise RuntimeError("Parameters cannot be changed after the "
                               "network has been sent to NEST!")
        self._neuron_param = value

    @Group.ids.setter
    def ids(self, value):
        if self._to_nest:
            raise RuntimeError("Ids cannot be changed after the "
                               "network has been sent to NEST!")

        self._ids = value

    @property
    def nest_gids(self):
        return self._nest_gids

    @property
    def has_model(self):
        return self._has_model

    @property
    def properties(self):
        dic = {
            "neuron_type": self.neuron_type,
            "neuron_model": self._neuron_model,
            "neuron_param": deepcopy(self._neuron_param)
        }
        return dic


class MetaNeuralGroup(MetaGroup, NeuralGroup):

    """
    Class defining a meta-group of neurons.

    Its main variables are:

    :ivar ~nngt.MetaGroup.ids: :obj:`list` of :obj:`int`
        the ids of the neurons in this group.
    :ivar ~nngt.MetaGroup.is_metagroup: :obj:`bool`
        whether the group is a meta-group or not (`neuron_type` is
        ``None`` for meta-groups)
    """

    def __init__(self, nodes=None, name=None, properties=None, **kwargs):
        '''
        Calling the class creates a group of neurons.
        The default is an empty group but it is not a valid object for
        most use cases.

        Parameters
        ----------
        nodes : int or array-like, optional (default: None)
            Desired size of the group or, a posteriori, NNGT indices of
            the neurons in an existing graph.
        name : str, optional (default: "Group N")
            Name of the meta-group.

        Returns
        -------
        A new :class:`~nngt.MetaNeuralGroup` object.
        '''
        kwargs["neuron_type"] = kwargs.get("neuron_type", None)

        super().__init__(nodes=nodes, name=name, properties=properties,
                         **kwargs)

    def __str__(self):
        return "MetaNeuralGroup({}size={})".format(
            self._name + ": " if self._name else "", self.size)

    @property
    def excitatory(self):
        '''
        Return the ids of all excitatory nodes inside the meta-group.
        '''
        if self.parent is not None:
            gtype = np.array(
                [g.neuron_type for g in self.parent.values()],
                dtype=int)

            ids = np.array(self.ids, dtype=int)

            parents = self.parent.get_group(ids, numbers=True)

            return ids[gtype[parents] == 1]

        return []

    @property
    def inhibitory(self):
        '''
        Return the ids of all inhibitory nodes inside the meta-group.
        '''
        if self.parent is not None:
            gtype = np.array(
                [g.neuron_type for g in self.parent.values()],
                dtype=int)

            ids = np.array(self.ids, dtype=int)

            parents = self.parent.get_group(ids, numbers=True)

            return ids[gtype[parents] == -1]

        return []

    @property
    def properties(self):
        return self._prop


class GroupProperty:

    """
    Class defining the properties needed to create groups of neurons from an
    existing :class:`~nngt.Graph` or one of its subclasses.

    :ivar ~nngt.GroupProperty.size: :obj:`int`
        Size of the group.
    :ivar constraints: :obj:`dict`, optional (default: {})
        Constraints to respect when building the
        :class:`~nngt.properties.NeuralGroup` .
    :ivar ~nngt.GroupProperty.neuron_model: str, optional (default: None)
        name of the model to use when simulating the activity of this group.
    :ivar ~nngt.GroupProperty.neuron_param: dict, optional (default: {})
        the parameters to use (if they differ from the model's defaults)
    """

    def __init__ (self, size, constraints={}, neuron_model=None,
                  neuron_param={}, syn_model=None, syn_param={}):
        '''
        Create a new instance of GroupProperties.

        Notes
        -----
        The constraints can be chosen among:
            - "avg_deg", "min_deg", "max_deg" (:class:`int`) to constrain the
              total degree of the nodes
            - "avg/min/max_in_deg", "avg/min/max_out_deg", to work with the
              in/out-degrees
            - "avg/min/max_betw" (:class:`double`) to constrain the betweenness
              centrality
            - "in_shape" (:class:`nngt.geometry.Shape`) to chose neurons inside
              a given spatial region

        Examples
        --------
        >>> di_constrain = { "avg_deg": 10, "min_betw": 0.001 }
        >>> group_prop = GroupProperties(200, constraints=di_constrain)
        '''
        self.size = size
        self.constraints = constraints
        self.neuron_model = neuron_model
        self.neuron_param = neuron_param
        self.syn_model = syn_model
        self.syn_param = syn_param


def _make_groups(graph, group_prop):
    '''
    Divide `graph` into groups using `group_prop`, a list of group properties
    @todo
    '''
    pass


# ----- #
# Tools #
# ----- #

def _check_syn_spec(syn_spec, group_names, groups):
    gsize = len(groups)
    # test if all types syn_spec are contained
    alltypes = set(((1, 1), (1, -1), (-1, 1), (-1, -1))).issubset(
        syn_spec.keys())
    # is there more than 1 type?
    types = list(set(g.neuron_type for g in groups))
    mt_type = len(types) > 1
    # check that only allowed entries are present
    edge_keys = []
    for k in syn_spec.keys():
        if isinstance(k, tuple):
            edge_keys.extend(k)
    edge_keys = set(edge_keys)
    allkeys = group_names + types
    assert edge_keys.issubset(allkeys), \
        '`syn_spec` edge entries can only be made from {}.'.format(allkeys)
    # warn if connections might be missing
    nspec = len(edge_keys)
    has_default = len(syn_spec) > nspec
    if mt_type and nspec < gsize**2 and not alltypes and not has_default:
        _log_message(
            logger, "WARNING",
            'There is not one synaptic specifier per inter-group'
            'connection in `syn_spec` and no default model was provided. '
            'Therefore, {} or 4 entries were expected but only {} were '
            'provided. It might be right, but make sure all cases are '
            'covered. Missing connections will be set as "static_'
            'synapse".'.format(gsize**2, nspec))
    for val in syn_spec.values():
        assert 'weight' not in val, '`weight` cannot be set here.'
        assert 'delay' not in val, '`delay` cannot be set here.'



"""
Utilities for the chord diagram.
"""

import numpy as np


def _get_normed_line(mat, i, x, start, end, is_sparse):
    if is_sparse:
        row = mat.getrow(i).todense().A1
        return (row / x[i]) * (end - start)

    return (mat[i, :] / x[i]) * (end - start)


def dist(points):
    x1, y1 = points[0]
    x2, y2 = points[1]

    return np.sqrt((x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1))


def polar2xy(r, theta):
    return np.array([r*np.cos(theta), r*np.sin(theta)])
